"""
CPU EMULATOR HANDLED INSTRUCTIONS

Add any instructions that need to be handled below.  The function should be declared as such

# Using the same function for multiple instructions:
@opcode("add")
@opcode("adc")
def _add(cpu_context, ip, mnem, operands):
    print "IN ADD"

# Using a single function for an opcode
@opcode
def MOV(cpu_context, ip, mnem, operands):
    print "IN MOV"

WARNING:
    Do NOT rely on the flags registers being correct.  There are places were flags are NOT being updated when they
    should, and the very fact that CALL instructions are skipped could cause flags to be incorrect.
"""

from __future__ import division

import logging
import re

import idautils
import idc

from .. import builtin_funcs
from .. import utils
from ..cpu_context import Operand
from ..registry import registrar

# Dictionary containing opcode names -> function
OPCODES = {}
opcode = registrar(OPCODES, name="opcode")


logger = logging.getLogger(__name__)


@opcode
def AAA(cpu_context, ip, mnem, operands):
    """ ASCII Adjust AX After Addition """
    if cpu_context.bitness == 64:
        logger.debug("{} 0x{:X} :: Opcode not valid for 64-bit".format(mnem, ip))
        return

    orig_ax = cpu_context.registers.ax

    if (cpu_context.registers.al & 0xF) > 9 or cpu_context.registers.af == 1:
        cpu_context.registers.ax += 0x106
        cpu_context.registers.af = 1
        cpu_context.registers.cf = 1
    else:
        cpu_context.registers.af = 0
        cpu_context.registers.cf = 0
    cpu_context.registers.al &= 0xF

    logger.debug("{} 0x{:X} :: Adjusted AX 0x{:X} -> 0x{:X}".format(mnem, ip, orig_ax, cpu_context.registers.ax))


@opcode
def AAD(cpu_context, ip, mnem, operands):
    """ ASCII Adjust AX Before Division """
    if cpu_context.bitness == 64:
        logger.debug("{} 0x{:X} :: Opcode not valid for 64-bit".format(mnem, ip))
        return

    orig_ax = cpu_context.registers.ax

    base = operands[0].value if operands else 10
    al = cpu_context.registers.al
    ah = cpu_context.registers.ah
    cpu_context.registers.al = (al + (ah * base)) & 0xFF
    cpu_context.registers.ah = 0

    logger.debug("{} 0x{:X} :: Adjusted AX 0x{:X} -> 0x{:X}".format(mnem, ip, orig_ax, cpu_context.registers.ax))


@opcode
def AAM(cpu_context, ip, mnem, operands):
    """ ASCII Adjust AX After Multiply """
    if cpu_context.bitness == 64:
        logger.debug("{} 0x{:X} :: Opcode not valid for 64-bit".format(mnem, ip))
        return

    orig_ax = cpu_context.registers.ax

    base = operands[0].value if operands else 10
    al = cpu_context.registers.al
    cpu_context.registers.ah = al // base
    cpu_context.registers.al = al % base

    logger.debug("{} 0x{:X} :: Adjusted AX 0x{:X} -> 0x{:X}".format(mnem, ip, orig_ax, cpu_context.registers.ax))


@opcode
def AAS(cpu_context, ip, mnem, operands):
    """ ASCII Adjust AX After Subtraction """
    if cpu_context.bitness == 64:
        logger.debug("{} 0x{:X} :: Opcode not valid for 64-bit".format(mnem, ip))
        return

    orig_ax = cpu_context.registers.ax

    if (cpu_context.registers.al & 0xF) > 9 or cpu_context.registers.af == 1:
        cpu_context.registers.ax -= 6
        cpu_context.registers.ah -= 1
        cpu_context.registers.af = 1
        cpu_context.registers.cf = 1
    else:
        cpu_context.registers.cf = 0
        cpu_context.registers.af = 0
    cpu_context.registers.al &= 0xF

    logger.debug("{} 0x{:X} :: Adjusted AX 0x{:X} -> 0x{:X}".format(mnem, ip, orig_ax, cpu_context.registers.ax))


@opcode("adc")
@opcode("add")
def _add(cpu_context, ip, mnem, operands):
    """
    Handle both ADC and ADD here since the only difference is the flags.
    """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    result = opvalue1 + opvalue2
    if mnem == "adc":
        result += cpu_context.registers.cf
    width = get_max_operand_size(operands)

    mask = utils.get_mask(width)
    cpu_context.registers.cf = int(result > mask)
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("{} 0x{:X} :: {} + {} = {}".format(mnem, ip, opvalue1, opvalue2, result))
    operands[0].value = result & mask


@opcode
def AND(cpu_context, ip, mnem, operands):
    """ AND logic operator """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    result = opvalue1 & opvalue2
    width = get_max_operand_size(operands)

    cpu_context.registers.cf = 0
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = 0
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("AND 0x{:X} :: {} & {} = {}".format(ip, opvalue1, opvalue2, result))
    operands[0].value = result


@opcode
def BSWAP(cpu_context, ip, mnem, operands):
    """ byte Swap """
    opvalue1 = operands[0].value
    width = operands[0].width
    result = swap_bytes(opvalue1, width)
    logger.debug("BSWAP 0x{:X} :: {} -> {}".format(ip, opvalue1, result))
    operands[0].value = result


def _sanitize_func_name(func_name):
    """Sanitizes the IDA function names to it's core name."""
    # remove the extra "_" IDA likes to add to the function name.
    if func_name.startswith("_"):
        func_name = func_name[1:]

    # Remove the numbered suffix IDA likes to add to duplicate function names.
    func_name = re.sub("_[0-9]+$", "", func_name)

    return func_name


@opcode
def CALL(cpu_context, ip, mnem, operands):
    """
    CALL function

    Attempt to determine the number of arguments passed to the function which are purged on return
    """
    # Function pointer can be a memory reference or immediate.
    func_ea = operands[0].addr or operands[0].value
    func_name = utils.get_function_name(func_ea)

    logger.debug("CALL 0x{:X} :: call {}".format(ip, func_name or "0x{:X}".format(func_ea)))

    # TODO: Should we be placing the ip on the stack to simulate the retn address?

    # If a valid function pointer, collect call history and emulate effects.
    if operands[0].is_func_ptr:
        args = cpu_context.get_function_args(func_ea)
        cpu_context.func_calls[ip] = (func_name, args)

        # Emulate the effects of any known builtin functions.
        func = builtin_funcs.get(func_ea)
        if not func:
            func = builtin_funcs.get(func_name)
            if not func:
                # Try one more time with a sanitized name.
                func_name = _sanitize_func_name(func_name)
                func = builtin_funcs.get(func_name)

        if func:
            try:
                logger.debug(" :: {}({})".format(func_name, ", ".join(map(repr, args))))
                ret = func(cpu_context, func_name, args)
                # Set return value to rax
                if ret is not None:
                    logger.debug(" :: Setting {!r} into rax".format(ret))
                    cpu_context.registers.rax = ret
            except RuntimeError:
                raise  # Allow RuntimeError exceptions to be thrown.
            except Exception as e:
                logger.debug(
                    "{:#08x} :: Failed to emulate builtin function: {}() with error: {}".format(ip, func_name, e)
                )

    if idc.__EA64__:
        return

    # For the called function, attempt to locate the function end and examine the "retn" instruction which
    # will contain the number of bytes to add back to SP.
    try:
        is_loaded = idc.is_loaded(func_ea)
    except TypeError:
        is_loaded = False

    # For non-loaded files, reset esp based on number of stack arguments.
    # (We are assuming the callee is responsible for resetting the stack)
    if not is_loaded:
        try:
            func_data = utils.get_function_data(func_ea)
            for arg in func_data:
                loc_type = arg.argloc.atype()
                # Where was this parameter passed?
                if loc_type == 1:  # ALOC_STACK
                    # reset esp for each stack argument.
                    cpu_context.registers.rsp += cpu_context.byteness
        except RuntimeError:
            # If the function data cannot be obtained, then adjust the stack based on the stack delta.
            sp_adjust = idc.get_sp_delta(idc.next_head(ip))
            # If sp_adjust is None, that means the next instruction is not in a function.
            # There is no way to determine the stack adjustment
            if sp_adjust is not None:
                cpu_context.registers.rsp += sp_adjust
    else:
        # Get address of retn instruction
        func_end = idc.get_func_attr(func_ea, idc.FUNCATTR_END)
        if func_end == idc.BADADDR:
            # If we can't get a valid function, then pull the stack adjustment using idc.get_sp_delta()
            # on the next instruction.
            sp_adjust = idc.get_sp_delta(idc.next_head(ip))
            # If sp_adjust is None, that means the next instruction is not in a function.
            # There is no way to determine the stack adjustment.
            if sp_adjust is not None:
                cpu_context.registers.rsp += sp_adjust

        else:
            # Find a "retn" and see if we need to adjust rsp.
            # All retn's should have the same operand so finding any of them will work.
            ea = func_end
            while ea > func_ea:
                if idc.print_insn_mnem(ea) == "retn":
                    sp_adjust = idc.get_operand_value(ea, 0)
                    # if retn doesn't adjust the stack, -1 is returned
                    if sp_adjust != -1:
                        cpu_context.registers.rsp += sp_adjust
                    return
                ea = idc.prev_head(ea)


@opcode
def CDQ(cpu_context, ip, mnem, operands):
    """ Convert DWORD to QWORD with sign extension """
    if cpu_context.registers.eax >> 31:
        result = 0xFFFFFFFF
    else:
        result = 0x0

    logger.debug("CDQ 0x{:X} :: Setting register EDX to 0x{:X}".format(ip, result))
    cpu_context.registers.edx = result


@opcode
def CLC(cpu_context, ip, mnem, operands):
    """ Clear Carry Flag """
    cpu_context.registers.cf = 0


@opcode
def CLD(cpu_context, ip, mnem, operands):
    """ Clear Direction Flag """
    cpu_context.registers.df = 0


@opcode
def CMC(cpu_context, ip, mnem, operands):
    """ Complement Carry Flag """
    cpu_context.registers.cf = int(not cpu_context.registers.cf)


@opcode
def CMP(cpu_context, ip, mnem, operands):
    """ Compare to values """
    width = get_min_operand_size(operands)
    mask = utils.get_mask(width)
    opvalue1 = operands[0].value & mask
    opvalue2 = operands[1].value & mask
    result = opvalue1 - opvalue2

    cpu_context.registers.cf = int((opvalue1 & mask) < (opvalue2 & mask))
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("CMP 0x{:X} :: {} <-> {} = {}".format(ip, opvalue1, opvalue2, result))


@opcode
def CMPS(cpu_context, ip, mnem, operands):
    """
    Nothing really to do for CMPS
    """
    pass


@opcode
def CMPSB(cpu_context, ip, mnem, operands):
    """
    TODO: Does this really need to be implemented for our purposes???
    """
    pass


@opcode
def CMPSW(cpu_context, ip, mnem, operands):
    """
    TODO: Does this really need to be implemented for our purposes???
    """
    pass


@opcode
def CMPSD(cpu_context, ip, mnem, operands):
    """
    TODO: Does this really need to be implemented for our purposes???
    """
    pass


@opcode
def CQO(cpu_context, ip, mnem, operands):
    """ Convert QWORD to DQWORD with sign extension """
    # Only works in 64-bit mode
    if cpu_context.bitness != 64:
        logger.debug("CQO 0x{:X} :: Opcode only available for 64-bit mode.")
        return

    if cpu_context.registers.rax >> 63:
        result = 0xFFFFFFFFFFFFFFFF
    else:
        result = 0x0

    logger.debug("CQO 0x{:X} :: Setting register RDX to 0x{:X}".format(ip, result))
    cpu_context.registers.rdx = result


@opcode
def CVTDQ2PD(cpu_context, ip, mnem, operands):
    """ Convert Packed Doubleword Integers to Packed Double-Precision Floating-Point Values """
    opvalue2 = operands[1].value
    dword0 = opvalue2 & 0xFFFFFFFF
    dword1 = (opvalue2 & 0xFFFFFFFF) >> 32
    dpfp0 = utils.float_to_int(dword0)
    dpfp1 = utils.float_to_int(dword1)
    result = (dpfp1 << 64) | dpfp0
    logger.debug("{} 0x{:X} :: {} -> {}, {} -> {} --> {}".format(mnem, ip, dword0, dpfp0, dword1, dpfp1, result))
    operands[0].value = result


@opcode
def CVTSI2SD(cpu_context, ip, mnem, operands):
    """ Convert Doubleword Int to Scalar Double-Precision Floating-Point """
    opvalue2 = operands[1].value
    result = utils.float_to_int(opvalue2)
    logger.debug("{} 0x{:X} :: int {} -> float equivalent {}".format(mnem, ip, opvalue2, result))
    operands[0].value = result


@opcode
def CVTTSD2SI(cpu_context, ip, mnem, operands):
    """ Convert with Truncation Scalar Double-Precision Floating-Point Value to Signed Integer """
    opvalue2 = operands[1].value
    # width = operands[0].width
    result = int(utils.int_to_float(opvalue2))
    logger.debug("{} 0x{:X} :: float {} -> int equivalent {}".format(mnem, ip, opvalue2, result))
    operands[0].value = result


@opcode
def CWD(cpu_context, ip, mnem, operands):
    """ Convert WORD to DWORD with sign extension """
    if cpu_context.registers.ax >> 15:
        result = 0xFFFF
    else:
        result = 0x0

    logger.debug("CWD 0x{:X} :: Setting register DX to 0x{:X}".format(ip, result))
    cpu_context.registers.dx = result


@opcode
def DEC(cpu_context, ip, mnem, operands):
    """ Decrement """
    opvalue1 = operands[0].value
    width = operands[0].width
    mask = utils.get_mask(width)
    result = opvalue1 - 1

    cpu_context.registers.af = int(result & 0x0F == 0x0F)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(utils.sign_bit(opvalue1, width) and not utils.sign_bit(result, width))
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["af", "zf", "sf", "of", "pf"], operands)

    logger.debug("DEC 0x{:X} :: {} - 1 = {}".format(ip, opvalue1, result))
    operands[0].value = result & mask


@opcode("div")
@opcode("idiv")
def DIV(cpu_context, ip, mnem, operands):
    """
    Divide

    rax / op1 -> rax (rdx holds remainder)
    """
    RAX_REG_SIZE_MAP = {8: "RAX", 4: "EAX", 2: "AX", 1: "AL"}
    RDX_REG_SIZE_MAP = {8: "RDX", 4: "EDX", 2: "DX"}

    divisor = operands[0].value
    width = operands[0].width
    if divisor == 0:
        # Log the instruction for a DIV / 0 error
        logger.debug("{} 0x{:X} :: DIV / 0".format(mnem, ip))
        return

    # We actually need to do some doctoring with DIV as operand 0 is implied as the EAX register of
    # a certain size.
    rax_str = RAX_REG_SIZE_MAP[width]
    dividend = cpu_context.registers[rax_str]

    result = (dividend // divisor) & utils.get_mask(width)
    remainder = (dividend % divisor) & utils.get_mask(width)
    logger.debug("DIV 0x{:X} :: {} / {} = {}".format(ip, dividend, divisor, result))
    if width == 1:
        # Result stored in AL, remainder stored in AH
        cpu_context.registers.al = result
        cpu_context.registers.ah = remainder
    else:
        rdx_str = RDX_REG_SIZE_MAP[width]
        cpu_context.registers[rax_str] = result
        cpu_context.registers[rdx_str] = remainder


@opcode
def DIVSD(cpu_context, ip, mnem, operands):
    """
    Divide Scalar Double-Precision Floating-Point Value

    op1 / op2 -> op1
    """
    opvalue1 = utils.int_to_float(operands[0].value)
    opvalue2 = utils.int_to_float(operands[1].value)
    # Because there is no guarantee that the registers/memory have been properly initialized, ignore DIV / 0 errors.
    if opvalue2 == 0:
        # Log DIV / 0 error
        logger.debug("{} 0x{:X} :: DIV / 0".format(mnem, ip))
        return

    result = opvalue1 // opvalue2
    logger.debug("{} 0x{:X} :: {} / {} = {}".format(mnem, ip, opvalue1, opvalue2, result))
    result = utils.float_to_int(result)
    operands[0].value = result


def _mul(cpu_context, ip, mnem, operands):
    """
    Handle MUL instruction and 1-operand IMUL instruction as the same.
    """
    RAX_REG_SIZE_MAP = {8: "RAX", 4: "EAX", 2: "AX", 1: "AL"}
    RDX_REG_SIZE_MAP = {8: "RDX", 4: "EDX", 2: "DX"}

    dx_reg = None
    dx_result = None
    width = get_max_operand_size(operands)
    mask = utils.get_mask(width)
    multiplier1 = cpu_context.reg_read(RAX_REG_SIZE_MAP[width])
    multiplier2 = operands[0].value
    result = multiplier1 * multiplier2
    flags = ["cf", "of"]

    if width == 1:
        ax_reg = RAX_REG_SIZE_MAP[2]
        ax_result = result
        if mnem.upper() == "MUL":
            cpu_context.registers.cf = 0
            cpu_context.registers.of = 0
    else:
        ax_reg = RAX_REG_SIZE_MAP[width]
        dx_reg = RDX_REG_SIZE_MAP[width]
        dx_result = (result & (utils.get_mask(width) << (width * 8))) >> (width * 8)
        ax_result = result & utils.get_mask(width)
        if mnem.upper() == "MUL":
            if result >> (width * 8):
                cpu_context.registers.cf = 1
                cpu_context.registers.of = 1
            else:
                cpu_context.registers.cf = 0
                cpu_context.registers.of = 0

    if mnem.upper() == "IMUL":
        cpu_context.registers.cf = int(
            not (
                (not utils.sign_bit(multiplier1, width) and multiplier2 & mask == 0)
                or (utils.sign_bit(multiplier1, width) and multiplier2 & mask == mask)
            )
        )
        cpu_context.registers.of = cpu_context.registers.cf
        cpu_context.registers.zf = int(multiplier1 & mask == 0)
        cpu_context.registers.sf = utils.sign_bit(multiplier1, width)
        cpu_context.registers.pf = get_parity(multiplier1)
        flags.extend(["zf", "sf", "pf"])

    cpu_context.jcccontext.update_flag_opnds(flags, operands)
    logger.debug(
        "{} 0x{:X} :: {} * {} = {} || EAX -> {} || EDX -> {}".format(
            mnem.upper(), ip, multiplier1, multiplier2, result, ax_result, dx_result if dx_reg else ""
        )
    )

    cpu_context.registers[ax_reg] = ax_result
    if dx_reg:
        cpu_context.registers[dx_reg] = dx_result


# TODO: Clean up mul, imul, and _mul
@opcode
def IMUL(cpu_context, ip, mnem, operands):
    """ Signed Multiplication

    ; Single operand form
    imul    ecx     ; Signed multiply the value in ecx with the value in eax (et.al)

    ; Two operand form
    imul    edi, edx    ; Signed multiply the destination operand (op 0) with the source operand (op 1)

    ; Three operand form
    imul    eax, edi, 5 ; Signed multiple source operand (op 1) with the immediate value (op 2) and store in
                        ; the destination operand (op 0)

    """
    width = get_max_operand_size(operands)
    op_count = len(operands)
    if op_count == 1:
        _mul(cpu_context, ip, mnem, operands)
        return
    elif op_count == 2:
        multiplier1 = operands[0].value
        multiplier2 = operands[1].value
    elif op_count == 3:
        multiplier1 = operands[1].value
        multiplier2 = operands[2].value
    else:
        raise Exception("0x{:X}: Invalid sequence for IMUL instruction".format(ip))

    mask = utils.get_mask(width)
    result = multiplier1 * multiplier2

    cpu_context.registers.cf = int(
        not (
            (not utils.sign_bit(multiplier1, width) and multiplier2 & mask == 0)
            or (utils.sign_bit(multiplier1, width) and multiplier2 & mask == mask)
        )
    )
    cpu_context.registers.of = cpu_context.registers.cf
    cpu_context.registers.zf = int(multiplier1 & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(multiplier1, width)
    cpu_context.registers.pf = get_parity(multiplier1)
    cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("IMUL 0x{:X} :: {} * {} = {}".format(ip, multiplier1, multiplier2, result))
    operands[0].value = result


@opcode
def INC(cpu_context, ip, mnem, operands):
    """ Increment """
    opvalue1 = operands[0].value

    result = opvalue1 + 1
    width = operands[0].width
    mask = utils.get_mask(width)

    logger.debug("INC 0x{:X} :: {} + 1 = {}".format(ip, opvalue1, result))
    operands[0].value = result

    cpu_context.registers.af = int(result & 0x0F == 0)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not utils.sign_bit(opvalue1, width) and utils.sign_bit(result, width))
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["af", "zf", "sf", "of", "pf"], operands)


@opcode
def JMP(cpu_context, ip, mnem, operands):
    """ Unconditional jump """
    jump_target = operands[0].value
    cpu_context.ip = jump_target


# TODO: Not all Jcc instructions are implemented here.
# TODO: Currently, these Jcc instructions assume that the instruction that modified the flags had 2 opcodes.
#   This is not always the case, (e.g. 'inc'). Also, we should probably figure out a way to move the
#   bulk of the code into a helper function.

# For the following jump instructions, the logic is basically the same
#   1. Get all the CodeRefs from the current IP (should only ever be 2)
#   2. Remove the EA that is the target of the Jcc instruction to we know where the non-jump target is
#   3. Determine the location where our condition takes us and set condition_target_ea to that
#   4. Set the value for the alternate path
# Note that since we aren't currently handling instructions which may cause conditional jumps, we need to
# determine if we have test_opnds and abort fixing the context if we don't.
@opcode("ja")
@opcode("jnbe")
def JA_JNBE(cpu_context, ip, mnem, operands):
    """ Jump Above (CF=0 && ZF=0) """
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition would take use based on our emulation and the value for the alt branch
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf", "zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.cf == 0 and cpu_context.registers.zf == 0:
        # opnd0 > opnd1 on this branch.  Set the alternate branch value opnd0 <= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 <= opnd1 on this branch. Set the alternate branch value opnd0 > opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jae")
@opcode("jnb")
@opcode("jnc")
def JAE_JNB(cpu_context, ip, mnem, operands):
    """ Jump Above or Equal / Jump Not Below / Jump Not Carry (CF=0) """
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.cf == 0:
        # opnd0 > opnd1 on this branch.  Set the alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 < opnd1 on this branch. Set the alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jb")
@opcode("jc")
@opcode("jnae")
def JB_JNAE(cpu_context, ip, mnem, operands):
    """ Jump Below / Jump Carry / Jump Not Above or Equal (CF=1) """
    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than the second operand of the compare operation.
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.cf:
        # opnd0 < opnd1 on this branch.  Set the alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd0 >= opnd1 on this branch. Set the alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jbe")
@opcode("jna")
def JBE_JNA(cpu_context, ip, mnem, operands):
    """ Jump Below or Equal / Jump Not Above (CF=1 || ZF=1) """
    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than or equal to the second operand of the compare operation.
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf", "zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.cf or cpu_context.registers.zf:
        # opnd0 <= opnd1 on this branch.  Set the alternate branch value opnd0 > opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd0 > opnd1 on this branch.  Set the alternate branch value opnd0 <= opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("je")
@opcode("jz")
def JE_JZ(cpu_context, ip, mnem, operands):
    """ Jump Equal / Jump Zero (ZF=1) """
    # Jump target contains the known data which is either 0 or the value of the second operand of the compare operation.
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    # There is additional logic that must be conducted for this jump.  If the src and dst operands are the same, then
    # the check was likely to determine if the value was 0 or not 0.  Else, the check was determining if src and dst
    # were equal.
    if cpu_context.registers.zf:
        # Indicates emulation likely produced two operands which were equal, or an operand which was 0, so set the
        # alternate branch to equal the value of operand 1 + 1.  Or the comparison was to check if an operand was 0.
        # eg:  cmp    eax, 0x3D     ; eax = 0x3D
        #      cmp    eax, 0x00     ; eax = 0
        #      test   rax, rax      ; rax = 0
        cpu_context.jcccontext.condition_target_ea = jump_target
        # There is no need to check if the operands are the same here.  If the check was for equality, then adding
        # 1 to the second operand will make the check not equal.  If the check was for 0 (both operands were the same),
        # simply adding 1 to 0 will make the test fail.
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # Indicates emulation likely produced two operands which were not equal, or an operand which was not 0, so set
        # the alternate branch to equal the value of operand 1.  Or the comparison was to check if an operand was 0.
        # eg:  cmp    eax, 0x3D     ; eax = 0
        #      cmp    eax, 0x00     ; eax = 7
        #      test   rax, rax      ; rax = 10
        cpu_context.jcccontext.condition_target_ea = next_inst
        # Need to determine if both operands are the same.  If they are, then the check was actually to determine
        # if the value was 0.  In this case, to make the test succeed, just set the value to 0.  Otherwise the test
        # was to compare two different operands for equality, so just set the first operand to the value of the second.
        if operand0.text == operand1.text:
            cpu_context.jcccontext.alt_branch_data = 0
        else:
            cpu_context.jcccontext.alt_branch_data = operand1.value

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jg")
@opcode("jnle")
def JG_JNLE(cpu_context, ip, mnem, operands):
    """ Jump Greater / Jump Not Less or Equal (ZF=0 && SF=OF) """
    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is larger than the second operand of the compare operation.
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf", "sf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.zf == 0 and cpu_context.registers.sf == cpu_context.registers.of:
        # opnd0 > opnd1 on this branch.  Set alternate branch value opnd0 <= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 <= opnd1on this branch.  Set alternate branch value opnd0 > opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jge")
@opcode("jnl")
def JGE_JNL(cpu_context, ip, mnem, operands):
    """ Jump Greater or Equal (SF=OF) """
    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is larger than or equal to the second operand of the compare operation.
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["sf", "of"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.sf == cpu_context.registers.of:
        # opnd0 >= opnd1 on this branch. Set alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 < opnd1 on this branch. Set alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jl")
@opcode("jnge")
def JL_JNGE(cpu_context, ip, mnem, operands):
    """ Jump Less (SF!=OF) """
    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than the second operand of the compare operation.
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["sf", "of"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.sf != cpu_context.registers.of:
        # opnd0 < opnd1 on this branch.  Set alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd1 >= opnd1 on this branch.  Set alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jle")
@opcode("jng")
def JLE_JNG(cpu_context, ip, mnem, operands):
    """ Jump Less or Equal (ZF=1 || SF!=OF) """
    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than or equal to the second operand of the compare operation.
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf", "sf", "of"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if cpu_context.registers.zf or cpu_context.registers.sf != cpu_context.registers.of:
        # opnd0 <= opnd2 on this branch.  Set alternate branch value opnd1 > opnd2
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd0 > opnd2 on this branch.  Set alternate branch value opnd0 <= opnd2
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{:X} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode("jne")
@opcode("jnz")
def JNE_JNZ(cpu_context, ip, mnem, operands):
    """ Jump Not Equal (ZF=0) """
    # Whatever the operation, it either set ZF or it didn't... Typically, the assumption can probably be made that
    # either the operands were equal such that a subtraction resulted in 0, or they weren't.
    # TODO: Does the compare instruction have an effect on which operand is the value to be used?
    jump_target = operands[0].value
    code_refs = list(idautils.CodeRefsFrom(ip, 1))
    code_refs.remove(jump_target)
    next_inst = code_refs[0]

    ## Set the target for which to modify the context, it will be the only address left in code_refs
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    # There is additional logic that must be conducted for this jump.  If the src and dst operands are the same, then
    # the check was likely determine if the value was 0 or not 0.  Else, the check was determining if src and dst were
    # not equal.
    if not cpu_context.registers.zf:
        # Indicates emulation likely produced two operands which were not equal, or an operand which was not 0, so set
        # the alternate branch to equal the value of operand 1 (which would also be 0).  Or the comparison was to check
        # if an operand was 0.
        # eg:  cmp    eax, 0x3D     ; eax = 0
        #      cmp    eax, 0x00     ; eax = 7
        #      test   rax, rax      ; rax = 10
        cpu_context.jcccontext.condition_target_ea = jump_target
        # Need to determine if the operands are the same operand (indicating a test for 0).  If the operands are the
        # same operand, then set the value to 0.  If the operands were not the same operand, the test was for
        # equality, so set the values to the same value.
        if operand0.text == operand1.text:
            cpu_context.jcccontext.alt_branch_data = 0
        else:
            cpu_context.jcccontext.alt_branch_data = operand1.value
    else:
        # Indicates emulation likely produced two operands which were equal, or an operand was 0, so set the
        # alternate branch to the value of operand 1 + 1.  Or the comparison was to check if an operand was 0.
        # eg:  cmp    eax, 0x00     ; eax = 0
        #      cmp    eax, 0x3D     ; eax = 0x3D
        #      test   rax, rax      ; rax = 0
        cpu_context.jcccontext.condition_target_ea = next_inst
        # There is no need to check if the operands were the same.  If the operands were the same, the test was to
        # determine if the operand was 0, so just adding 1 to the operand will be enough to create the false condition.
        # If the operands weren't the same, the check was to determine if the operands were equal, so adding 1 to
        # the second operand will be enough to make the condition false.
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "{} 0x{:X} :: Primary branch 0x{:X}, using value 0x{} for alternate branch".format(
            mnem, ip, cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )
    )


@opcode
def JNO(cpu_context, ip, mnem, operands):
    """ Jump Not Overflow (OF=0) """
    pass


@opcode("jnp")
@opcode("jpo")
def JNP_JPO(cpu_context, ip, mnem, operands):
    """ Jump Not Parity (PF=0) """
    pass


@opcode
def JNS(cpu_context, ip, mnem, operands):
    """ Jump Not Sign (SF=0) """
    pass


@opcode
def JO(cpu_context, ip, mnem, operands):
    """ Jump Overflow (OF=1) """
    pass


@opcode("jp")
@opcode("jpe")
def JP_JPE(cpu_context, ip, mnem, operands):
    """ Jump Parity (PF=1) """
    pass


@opcode
def JS(cpu_context, ip, mnem, operands):
    """ Jump Sign (SF=1) """
    pass


@opcode
def LEA(cpu_context, ip, mnem, operands):
    """
    Handle the LEA instruction.
    """
    address = operands[1].addr
    logger.debug("{} 0x{:X} :: Copy address 0x{:X} into {}".format(mnem, ip, address, operands[0].text))
    operands[0].value = address


@opcode("mov")
@opcode("movzx")
@opcode("movapd")
@opcode("movaps")
@opcode("movdqa")
@opcode("movdqu")
@opcode("movupd")
@opcode("movups")
def _mov(cpu_context, ip, mnem, operands):
    """
    Handle the MOV, MOVZX, MOVA*, MOVD*, MOVU* instructions in the same manner.

    MOVZX is a zero extend, but this logic makes no real sense in python.

    NOTE: Since the widths are already taken into account when the operand values are retrieved
    or set, the logic for most mov* instructions are the same.
    """
    opvalue2 = operands[1].value
    logger.debug("{} 0x{:X} :: Copy {} into {}".format(mnem, ip, opvalue2, operands[0].text))
    operands[0].value = opvalue2


@opcode("movsx")
@opcode("movsxd")
def _movsx(cpu_context, ip, mnem, operands):
    """ Move with Sign Extend """
    opvalue2 = operands[1].value
    logger.debug("MOVSX 0x{:X} :: Sign-extend {} into {}".format(ip, opvalue2, operands[0].text))
    size = utils.sign_extend(opvalue2, operands[1].width, operands[0].width)
    operands[0].value = size


@opcode("movs")  # I don't believe IDA will ever use just "movs", but it's here just incase.
@opcode("movsb")
@opcode("movsw")
@opcode("movsd")
def movs(cpu_context, ip, mnem, operands):
    """
    Move Scalar Double-Precision Floating-Point Value
    OR
    Move Data from String to String
    """
    # movsd op1 op2
    if mnem == "movsd" and len(operands) == 2:
        op1, op2 = operands
        data = op2.value
        if op1.is_register:
            # When moving into an XMM register, the high 64 bits needs to remain untouched.
            data = (data & 0xFFFFFFFFFFFFFFFF0000000000000000) | data
        logger.debug("{} 0x{:X} :: {} -> {}".format(mnem, ip, op2.value, data))
        op1.value = data

    # movs*
    else:
        if cpu_context.bitness == 16:
            src = "SI"
            dst = "DI"
        else:
            src = "ESI"
            dst = "EDI"
        # IDA sometimes provides a single "fake" operand to help determine the size.
        width = operands[0].width if operands else 4

        size = {"movs": width, "movsb": 1, "movsw": 2, "movsd": 4}[mnem]
        src_ptr = cpu_context.registers[src]
        dst_ptr = cpu_context.registers[dst]
        logger.debug("{} 0x{:X} :: 0x{:X} -> 0x{:X}".format(mnem, ip, src_ptr, dst_ptr))
        cpu_context.mem_copy(src_ptr, dst_ptr, size)

        # update ESI/EDI registers
        if cpu_context.registers.df:
            cpu_context.registers[src] -= size
            cpu_context.registers[dst] -= size
        else:
            cpu_context.registers[src] += size
            cpu_context.registers[dst] += size


@opcode
def MOVD(cpu_context, ip, mnem, operands):
    """ Move Dword """
    opvalue2 = operands[1].value & 0xFFFFFFFF
    logger.debug("{} 0x{:X} :: Copy {} into {}".format(mnem, ip, opvalue2, operands[0].text))
    operands[0].value = opvalue2


@opcode
def MOVQ(cpu_context, ip, mnem, operands):
    """ Move Quadword """
    opvalue2 = operands[1].value & 0xFFFFFFFFFFFFFFFF
    logger.debug("{} 0x{:X} :: Copy {} into {}".format(mnem, ip, opvalue2, operands[0].text))
    operands[0].value = opvalue2


@opcode
def MUL(cpu_context, ip, mnem, operands):
    """ Multiplication """
    _mul(cpu_context, ip, mnem, operands)


@opcode
def NEG(cpu_context, ip, mnem, operands):
    """ Negate """
    opvalue1 = operands[0].value
    result = -opvalue1
    width = operands[0].width
    mask = utils.get_mask(width)

    cpu_context.registers.cf = int(result & mask != 0)
    cpu_context.registers.af = int(result & 0x0F != 0)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(utils.sign_bit(opvalue1, width) and not utils.sign_bit(result, width))
    cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of"], operands)

    logger.debug("NEG 0x{:X} :: {} - {}".format(ip, opvalue1, result))
    operands[0].value = result


@opcode
def NOT(cpu_context, ip, mnem, operands):
    """ NOT Logic Operator """
    opvalue1 = operands[0].value
    result = ~opvalue1
    logger.debug("NOT 0x{:X} :: {} -> {}".format(ip, opvalue1, result))
    operands[0].value = result


@opcode
def OR(cpu_context, ip, mnem, operands):
    """ OR Logic Operator """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 | opvalue2

    cpu_context.registers.cf = 0
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = 0
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("OR 0x{:X} :: {} | {} = {}".format(ip, opvalue1, opvalue2, result))
    operands[0].value = result


@opcode
def POP(cpu_context, ip, mnem, operands):
    """ POP stack value """
    result = utils.struct_unpack(cpu_context.mem_read(cpu_context.sp, cpu_context.byteness))
    cpu_context.sp += cpu_context.byteness
    logger.debug("POP 0x{:X} :: Popped value {} into {}".format(ip, result, operands[0].text))
    operands[0].value = result


@opcode("popa")
@opcode("popad")
def POPA(cpu_context, ip, mnem, operands):
    """
    POPA (valid only for x86)

    NOTE: This function will return None.  This is one instance where accessing the registers directly makes more
            sense.
    """
    # NOTE Some assemblers may force size based on operand size instead of mnem.
    # However, IDA should set the proper mnemonic for us.
    if mnem.endswith("d"):
        reg_order = ["EDI", "ESI", "EBP", "ESP", "EBX", "EDX", "ECX", "EAX"]
    else:
        reg_order = ["DI", "SI", "BP", "SP", "BX", "DX", "CX", "AX"]

    logger.debug("{} 0x{:X}".format(mnem, ip))
    for reg in reg_order:
        if reg not in ("ESP", "SP"):
            # reg <- Pop()
            val = utils.struct_unpack(cpu_context.mem_read(cpu_context.registers.esp, cpu_context.byteness))
            cpu_context.registers[reg] = val
            logger.debug("{}} 0x{:X} :: Popped value {} into {}".format(mnem, ip, val, reg))
        cpu_context.sp += cpu_context.byteness


@opcode("popf")
@opcode("popfd")
@opcode("popfq")
def POPF(cpu_context, ip, mnem, operands):
    """ Pop FLAGS/EFLAGS register off the stack """
    flags = utils.struct_unpack(cpu_context.mem_read(cpu_context.sp, cpu_context.byteness))
    cpu_context.sp += cpu_context.byteness

    logger.debug("{} 0x{:X} :: Popped value {} into flags register".format(mnem, ip, flags))
    if cpu_context.bitness == 16:
        cpu_context.registers.flags = flags
    else:
        # Also works for RFLAGS. Since we don't support them, they are all zeros.
        cpu_context.registers.eflags = flags


@opcode
def PUSH(cpu_context, ip, mnem, operands):
    """ PUSH """
    operand = operands[0]
    logger.debug("PUSH 0x{:X} :: Pushing {} onto stack".format(ip, operand.value))
    cpu_context.registers.rsp -= cpu_context.byteness
    cpu_context.mem_write(cpu_context.registers.esp, utils.struct_pack(operand.value, width=operand.width))


@opcode("pusha")
@opcode("pushad")
def PUSHA(cpu_context, ip, mnem, operands):
    """ Push all general-purpose registers (valid only for x86) """
    # NOTE Some assemblers may force size based on operand size instead of mnem.
    # However, IDA should set the proper mnemonic for us.
    if mnem.endswith("d"):
        reg_order = ["EAX", "ECX", "EDX", "EBX", "ESP", "EBP", "ESI", "EDI"]
        orig_esp = cpu_context.registers.esp
    else:
        reg_order = ["AX", "CX", "DX", "BX", "SP", "BP", "SI", "DI"]
        orig_esp = cpu_context.registers.sp

    logger.debug("{} 0x{:X}".format(mnem, ip))
    for reg in reg_order:
        cpu_context.sp -= cpu_context.byteness
        pushed_value = orig_esp if reg in ("ESP", "SP") else cpu_context.registers[reg]
        logger.debug("{} 0x{:X} :: Pushing {} onto stack".format(mnem, ip, pushed_value))
        cpu_context.mem_write(cpu_context.registers.esp, utils.struct_pack(pushed_value))


@opcode("pushf")
@opcode("pushfd")
@opcode("pushfq")
def PUSHF(cpu_context, ip, mnem, operands):
    """ Push FLAGS/EFLAGS register onto the stack """
    if cpu_context.bitness == 16:
        flags = cpu_context.registers.flags
    else:
        # Also works for RFLAGS. Since we don't support them, they are all zeros.
        flags = cpu_context.registers.eflags

    # VM and RF flags are not copied.
    flags &= ~0x10000  # rf
    flags &= ~0x20000  # vm

    logger.debug("{} 0x{:X} :: Pushing {} onto the stack".format(mnem, ip, flags))
    cpu_context.sp -= cpu_context.byteness
    cpu_context.mem_write(cpu_context.sp, utils.struct_pack(flags))


@opcode
def RCR(cpu_context, ip, mnem, operands):
    """ Rotate Carry Right """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 9, 17, 33 when the width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % ((width * 8) + 1)
    else:
        # This is undefined behavior
        return

    if opvalue2 == 1:
        cpu_context.registers.of = get_msb(opvalue1, width) ^ cpu_context.registers.cf

    while tempcount:
        tempcf = get_lsb(opvalue2)
        opvalue1 = (opvalue1 >> 1) + (cpu_context.registers.cf * 2 ** width)
        cpu_context.registers.cf = tempcf
        tempcount -= 1

    cpu_context.jcccontext.update_flag_opnds(["cf"], operands)
    logger.debug("RCR 0x{:X} :: Rotate {} right by {} -> {}".format(ip, operands[0].value, opvalue2, opvalue1))
    operands[0].value = opvalue1


@opcode
def RCL(cpu_context, ip, mnem, operands):
    """ Rotate Carry Left """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 9, 17, 33 when width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % ((width * 8) + 1)
    else:
        # This is undefined behavior
        return

    while tempcount:
        tempcf = get_msb(opvalue1, width)
        opvalue1 = (opvalue1 * 2) + cpu_context.registers.cf
        cpu_context.registers.cf = tempcf
        tempcount -= 1

    if opvalue2 == 1:
        cpu_context.registers.of = get_msb(opvalue1, width) ^ cpu_context.registers.cf

    cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("RCL 0x{:X} :: Rotate {} left by {} -> {}".format(ip, operands[0].value, opvalue2, opvalue1))
    operands[0].value = opvalue1


@opcode
def ROL(cpu_context, ip, mnem, operands):
    """ Rotate Left """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 8, 16, 32 when width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % (width * 8)
    else:
        # This is undefined behavior
        return

    if tempcount > 0:
        while tempcount:
            tempcf = get_msb(opvalue1, width)
            opvalue1 = (opvalue1 * 2) + tempcf
            tempcount -= 1

        cpu_context.registers.cf = get_lsb(opvalue1)
        if opvalue2 == 1:
            cpu_context.registers.of = get_msb(opvalue1, width) ^ cpu_context.registers.cf

    cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("ROL 0x{:X} :: Rotate {} left by {} -> {}".format(ip, operands[0].value, opvalue2, opvalue1))
    operands[0].value = opvalue1


@opcode
def ROR(cpu_context, ip, mnem, operands):
    """ Rotate Right """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 8, 16, 32 when width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % (width * 8)
    else:
        # This is undefined behavior
        return

    if tempcount > 0:
        while tempcount:
            tempcf = get_lsb(opvalue2)
            opvalue1 = (opvalue1 >> 1) + (tempcf * 2 ** width)
            tempcount -= 1

        cpu_context.registers.cf = get_msb(opvalue1, width)
        if opvalue2 == 1:
            cpu_context.registers.of = get_msb(opvalue1, width) ^ (get_msb(opvalue1, width) - 1)

    cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("ROR 0x{:X} :: Rotate {} right by {} -> {}".format(ip, operands[0].value, opvalue2, opvalue1))
    operands[0].value = opvalue1


@opcode("sal")
@opcode("shl")
def sal_shl(cpu_context, ip, mnem, operands):
    """ Shift Arithmetic Left """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1

    if opvalue2:
        # 0x3F Because we want to allow for 64-bit code
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
        while tempcount:
            cpu_context.registers.cf = get_msb(result, width)
            result *= 2
            tempcount -= 1

        result &= utils.get_mask(width)

        bit_count = width * 8
        if opvalue2 <= bit_count:
            cpu_context.registers.cf = (opvalue1 >> (bit_count - opvalue2)) & 0x01
        else:
            cpu_context.registers.cf = 0
        cpu_context.registers.sf = utils.sign_bit(result, width)
        if opvalue2 == 1:
            cpu_context.registers.of = utils.sign_bit(opvalue1 ^ result, width)
        else:
            cpu_context.registers.of = utils.sign_bit((opvalue1 << (opvalue2 - 1)) ^ result, width)

    cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("SAL 0x{:X} :: Shift {} left by {} -> {}".format(ip, opvalue1, opvalue2, result))
    operands[0].value = result


@opcode
def SAR(cpu_context, ip, mnem, operands):
    """ Shift Arithmetic Right """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1
    if opvalue2:
        # 0x3F Because we want to allow for 64-bit code
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
        msb = get_msb(opvalue1, cpu_context.byteness)
        while tempcount:
            cpu_context.registers.cf = get_lsb(result)
            result = result >> 1
            tempcount -= 1

        bit_count = width * 8
        if opvalue2 < bit_count:
            cpu_context.registers.cf = (opvalue1 >> (opvalue2 - 1)) & 0x01
        else:
            cpu_context.registers.cf = utils.sign_bit(opvalue1, width)
        cpu_context.registers.zf = int(result == 0)
        cpu_context.registers.sf = utils.sign_bit(result, width)
        cpu_context.registers.of = 0
        cpu_context.registers.pf = get_parity(result)

        result |= msb << cpu_context.bitness

    cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)
    logger.debug("SAR 0x{:X} :: Shift {} right by {} -> {}".format(ip, opvalue1, opvalue2, result))
    operands[0].value = result


@opcode
def SBB(cpu_context, ip, mnem, operands):
    """ Subtract with Borrow/Carry """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    mask = utils.get_mask(width)
    result = opvalue1 - (opvalue2 + cpu_context.registers.cf)

    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["af", "zf", "sf", "of", "pf"], operands)
    logger.debug("SBB 0x{:X} :: {} - {} = {}".format(ip, opvalue1, (opvalue2 + cpu_context.registers.cf), result))
    operands[0].value = result


# TODO: Do we need SCAS* implemented????


@opcode
def SETNA(cpu_context, ip, mnem, operands):
    """ Set if Not Above """
    result = int(cpu_context.registers.zf or cpu_context.registers.cf)
    logger.debug("SETNA 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETLE(cpu_context, ip, mnem, operands):
    """ Set if Less than or Equal """
    result = int(cpu_context.registers.zf or (cpu_context.registers.sf != cpu_context.registers.of))
    logger.debug("SETLE 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETGE(cpu_context, ip, mnem, operands):
    """ Set if Greater than or Equal """
    result = int(cpu_context.registers.sf == cpu_context.registers.of)
    logger.debug("SETGE 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETG(cpu_context, ip, mnem, operands):
    """ Set if Greather than """
    result = int(cpu_context.registers.zf and (cpu_context.registers.sf == cpu_context.registers.of))
    logger.debug("{} 0x{:X} :: Setting {} to {}".format(mnem.upper(), ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETE(cpu_context, ip, mnem, operands):
    """ Set if Equal """
    result = int(cpu_context.registers.zf)
    logger.debug("{} 0x{:X} :: Setting {} to {}".format(mnem.upper(), ip, operands[0].text, result))
    operands[0].value = result
    # cpu_context.set_operand_value(0, result)


@opcode
def SETC(cpu_context, ip, mnem, operands):
    """ Set if Carry """
    result = int(cpu_context.registers.cf)
    logger.debug("{} 0x{:X} :: Setting {} to {}".format(mnem.upper(), ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETBE(cpu_context, ip, mnem, operands):
    """ Set if Below or Equal """
    result = int(cpu_context.registers.cf and cpu_context.registers.zf)
    logger.debug("SETBE 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETB(cpu_context, ip, mnem, operands):
    """ Set if Below """
    SETC(cpu_context, ip, mnem, operands)


@opcode
def SETAE(cpu_context, ip, mnem, operands):
    """ Set if Above or Equal """
    SETC(cpu_context, ip, mnem, operands)


@opcode
def SETA(cpu_context, ip, mnem, operands):
    """ Set if Above """
    result = int(not (cpu_context.registers.cf | cpu_context.registers.zf))
    logger.debug("SETA 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETPS(cpu_context, ip, mnem, operands):
    """ Set if Not??? Parity """
    result = int(cpu_context.registers.sf)
    logger.debug("SETPS 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETPO(cpu_context, ip, mnem, operands):
    """ Set if Parity Odd """
    result = int(cpu_context.registers.pf)
    logger.debug("{} 0x{:X} :: Setting {} to {}".format(mnem.upper(), ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETPE(cpu_context, ip, mnem, operands):
    """ Set if Parity Event """
    SETPO(cpu_context, ip, mnem, operands)


@opcode
def SETP(cpu_context, ip, mnem, operands):
    """ Set if Parity """
    SETPO(cpu_context, ip, mnem, operands)


@opcode
def SETO(cpu_context, ip, mnem, operands):
    """ Set if Overflow """
    result = int(cpu_context.registers.of)
    logger.debug("SETO 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNS(cpu_context, ip, mnem, operands):
    """ Set if Not Sign """
    result = int(not cpu_context.registers.sf)
    logger.debug("SETNS 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNP(cpu_context, ip, mnem, operands):
    """ Set if Not Parity """
    result = int(not cpu_context.registers.pf)
    logger.debug("SETNP 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNO(cpu_context, ip, mnem, operands):
    """ Set if Not Overflow """
    result = int(not cpu_context.registers.of)
    logger.debug("SETNO 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNL(cpu_context, ip, mnem, operands):
    """ Set if Not Less """
    result = int(cpu_context.registers.sf == cpu_context.registers.of)
    logger.debug("SETNL 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result
    # cpu_context.set_operand_value(0, result)


@opcode
def SETNGE(cpu_context, ip, mnem, operands):
    """ Set if Not Greater Than or Equal """
    result = int(cpu_context.registers.sf != cpu_context.registers.of)
    logger.debug("{} 0x{:X} :: Setting {} to {}".format(mnem.upper(), ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNG(cpu_context, ip, mnem, operands):
    """ Set if Not Greater """
    result = int(cpu_context.registers.zf or (cpu_context.registers.sf != cpu_context.registers.of))
    logger.debug("SETNG 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNE(cpu_context, ip, mnem, operands):
    """ Set if Not Equal """
    result = int(not cpu_context.registers.zf)
    logger.debug("{} 0x{:X} :: Setting {} to {}".format(mnem.upper(), ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNC(cpu_context, ip, mnem, operands):
    """ Set if Not Carry """
    result = int(not cpu_context.registers.cf)
    logger.debug("{} 0x{:X} :: Setting {} to {}".format(mnem.upper(), ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNBE(cpu_context, ip, mnem, operands):
    """ Set if Not Below or Equal """
    result = int(not (cpu_context.registers.cf | cpu_context.registers.zf))
    logger.debug("SETNBE 0x{:X} :: Setting {} to {}".format(ip, operands[0].text, result))
    operands[0].value = result


@opcode
def SETNB(cpu_context, ip, mnem, operands):
    """ Set if Not Below """
    SETNC(cpu_context, ip, mnem, operands)


@opcode
def SETNAE(cpu_context, ip, mnem, operands):
    """ Set if Not Above or Equal """
    SETC(cpu_context, ip, mnem, operands)


@opcode
def SETL(cpu_context, ip, mnem, operands):
    """ Set if Less Than """
    SETNGE(cpu_context, ip, mnem, operands)


@opcode
def SETNLE(cpu_context, ip, mnem, operands):
    """ Set if Not Less Than or Equal """
    SETG(cpu_context, ip, mnem, operands)


@opcode
def SETNZ(cpu_context, ip, mnem, operands):
    """ Set if Not Zero """
    SETNE(cpu_context, ip, mnem, operands)


@opcode
def SETZ(cpu_context, ip, mnem, operands):
    """ Set if Zero """
    SETE(cpu_context, ip, mnem, operands)


@opcode
def SHR(cpu_context, ip, mnem, operands):
    """ Shift Right """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    if not opvalue2:
        return

    result = opvalue1
    tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    while tempcount:
        cpu_context.registers.cf = get_lsb(result)
        result >>= 1
        tempcount -= 1

    cpu_context.registers.cf = (opvalue1 >> (opvalue2 - 1)) & 0x01
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    if opvalue2 == 1:
        cpu_context.registers.of = utils.sign_bit(opvalue1, width)
    else:
        cpu_context.registers.of = 0

    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)
    logger.debug("SHR 0x{:X} :: Shift {} right by {} -> {}".format(ip, opvalue1, opvalue2, result))
    operands[0].value = result


@opcode("stosb")
@opcode("stosw")
@opcode("stosd")
@opcode("stosq")
def STOSx(cpu_context, ip, mnem, operands):
    """ STOre value in {R,E}AX, AX, AL in the address pointed to by {R,E}DI"""
    # Make a mapping for the opcode to define what registers we are working with.
    RAX_REG_SIZE_MAP = {8: "RAX", 4: "EAX", 2: "AX", 1: "AL"}
    RDI_REG_SIZE_MAP = {8: "RDI", 4: "EDI"}

    # Recreate the dst and src operands, since IDA hides them.
    # (We can't use operands, because they are fake and hidden.)
    dst_opnd = Operand(cpu_context, ip, 0)
    dst_opnd.text = RDI_REG_SIZE_MAP.get(dst_opnd.width, "EDI")
    dst_opnd.type = idc.o_reg
    src_opnd = Operand(cpu_context, ip, 1)
    src_opnd.text = RAX_REG_SIZE_MAP[src_opnd.width]
    src_opnd.type = idc.o_reg

    logger.debug("{} 0x{:X} :: Storing {} at 0x{}".format(mnem, ip, src_opnd.value, dst_opnd.value))
    data = utils.struct_pack(src_opnd.value, width=src_opnd.width)
    cpu_context.mem_write(dst_opnd.value, data)
    if cpu_context.registers.df:
        dst_opnd.value -= src_opnd.width
    else:
        dst_opnd.value += src_opnd.width


@opcode
def SUB(cpu_context, ip, mnem, operands):
    """ Subtract """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 - opvalue2

    mask = utils.get_mask(width)
    cpu_context.registers.cf = int((opvalue1 & mask) < (opvalue2 & mask))
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("SUB 0x{:X} :: {} - {} = {}".format(ip, opvalue1, opvalue2, result))
    operands[0].value = result


@opcode
def TEST(cpu_context, ip, mnem, operands):
    """ Test values for equality """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 & opvalue2

    mask = utils.get_mask(width)
    cpu_context.registers.cf = int((opvalue1 & mask) < (opvalue2 & mask))
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("TEST 0x{:X} :: {} & {} -> {}".format(ip, opvalue1, opvalue2, result))


@opcode
def XCHG(cpu_context, ip, mnem, operands):
    """ Exchange two values """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    logger.debug("XCHG 0x{:X} :: exchange {} and {}".format(ip, opvalue1, opvalue2))
    operands[1].value = opvalue1
    operands[0].value = opvalue2


@opcode("xor")
@opcode("pxor")
def _xor(cpu_context, ip, mnem, operands):
    """ XOR """
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 ^ opvalue2

    cpu_context.registers.cf = 0
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = 0
    cpu_context.registers.pf = get_parity(result)
    cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("{} 0x{:X} :: {} ^ {} = {}".format(mnem, ip, opvalue1, opvalue2, result))
    operands[0].value = result


# Global helper functions


def get_max_operand_size(operands):
    """
    Given the list of named tuples containing the operand value and bit width, determine the largest bit width.

    :param operands: list of Operand objects

    :return: largest operand width
    """
    return max(operand.width for operand in operands)


def get_min_operand_size(operands):
    """
    Given the list of named tuples containing the operand value and bit width, determine the smallest bit width.

    :param operands: list of Operand objects

    :return: smallest operand width
    """
    return min(operand.width for operand in operands)


def get_msb(value, size):
    """
    Get most significant bit.

    :param value: value to obtain msb from

    :size: bit width of value in bytes

    :return: most significant bit
    """
    return value >> ((8 * size) - 1)


def get_lsb(value):
    """
    Get least significant bit.

    :param value: value to obtain lsb from

    :return: least significant bit
    """
    return value & 0x1


def swap_bytes(value, size):
    """
    Swaps a set of bytes based on size

    :param value: value to swap

    :param size: width of value in bytes

    :return: swapped
    """
    if size == 1:
        return value

    if size == 2:
        return ((value & 0xFF) << 8) | ((value & 0xFF00) >> 8)

    if size == 4:
        return (
            ((value & 0xFF) << 24)
            | (((value & 0xFF00) >> 8) << 16)
            | (((value & 0xFF0000) >> 16) << 8)
            | ((value & 0xFF000000) >> 24)
        )

    if size == 8:
        return (
            ((value & 0xFF) << 56)
            | (((value & 0xFF00) >> 8) << 48)
            | (((value & 0xFF0000) >> 16) << 40)
            | (((value & 0xFF000000) >> 24) << 32)
            | (((value & 0xFF00000000) >> 32) << 24)
            | (((value & 0xFF0000000000) >> 40) << 16)
            | (((value & 0xFF000000000000) >> 48) << 8)
            | ((value & 0xFF00000000000000) >> 56)
        )


# fmt: off
parity_lookup_table = [
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
]
# fmt: on


def get_parity(value):
    """Returns the parity of the given value."""
    return parity_lookup_table[value & 0xFF]



"""
Components for emulating an x86/x64 architecture.
"""

from ..cpu_context import ProcessorContext
from ..registry import registrar
from .registers import ARM_Registers
from .opcodes import OPCODES


STACK_BASE = 0x1180000  # Base address for stack

RSP_OFFSET = 0x800
RBP_OFFSET = 0x400


class ARMProcessorContext(ProcessorContext):
    """Processor context for ARM architecture"""

    ARCH_NAME = "ARM"
    OPCODES = OPCODES

    def __init__(self):
        super(ARMProcessorContext, self).__init__(
            ARM_Registers(), instruction_pointer="pc", stack_pointer="sp", stack_registers=["sp", "wsp"],
        )
        # TODO: Set up stack correctly for ARM
        # # Set up the stack before we go.
        # self.registers.rsp = STACK_BASE - RSP_OFFSET
        # self.registers.rbp = STACK_BASE - RBP_OFFSET



"""
Components for emulating an x86/x64 architecture.
"""

from copy import deepcopy

from ..cpu_context import ProcessorContext
from .registers import x86_64_Registers
from .opcodes import OPCODES
from . import fpu_opcodes  # trigger registration


STACK_BASE = 0x1180000  # Base address for stack

RSP_OFFSET = 0x800
RBP_OFFSET = 0x400


class x86_64ProcessorContext(ProcessorContext):
    """Processor context for x86/x64 architecture"""

    ARCH_NAME = "metapc"
    OPCODES = OPCODES

    def __init__(self):
        super(x86_64ProcessorContext, self).__init__(
            x86_64_Registers(),
            instruction_pointer="rip",
            stack_pointer="rsp",
            stack_registers=["rsp", "esp", "rbp", "ebp"],
        )
        # Set up the stack before we go.
        self.registers.rsp = STACK_BASE - RSP_OFFSET
        self.registers.rbp = STACK_BASE - RBP_OFFSET



"""
FPU instructions.
"""

import logging
import operator

from .opcodes import opcode
from .. import utils


logger = logging.getLogger(__name__)


# TODO: Handle stack fault exceptions.
# TODO: Handle overflow and division by zero exceptions.
# TODO: Support proper rounding support and REAL4, REAL8, REAL10 data types.
# TODO: Add missing opcodes.


@opcode("fadd")
@opcode("fiadd")
@opcode("faddp")
@opcode("fdiv")
@opcode("fdivr")
@opcode("fidiv")
@opcode("fidivr")
@opcode("fdivp")
@opcode("fdivrp")
@opcode("fmul")
@opcode("fimul")
@opcode("fmulp")
@opcode("fsub")
@opcode("fsubr")
@opcode("fisub")
@opcode("fisubr")
@opcode("fsubp")
@opcode("fsubrp")
def _compute(cpu_context, ip, mnem, operands):
    """Perform add/sub/mul/div computation on floating point numbers."""
    # Determine operator.
    if "add" in mnem:
        op, op_str = operator.add, "+"
    elif "sub" in mnem:
        op, op_str = operator.sub, "-"
    elif "div" in mnem:
        op, op_str = operator.div, "/"
    elif "mul" in mnem:
        op, op_str = operator.mul, "*"
    else:
        raise RuntimeError("Invalid mnem: {}".format(mnem))

    # Collect terms.
    if not operands:
        term1 = cpu_context.registers.st0
        term2 = cpu_context.registers.st1
    elif len(operands) == 1:
        term1 = cpu_context.registers.st0
        term2 = operands[0].value
    elif len(operands) == 2:
        term1 = operands[0].value
        term2 = operands[1].value
    else:
        logger.debug("{} 0x{:X} :: Unexpected number of operands: {}".format(mnem, ip, len(operands)))
        return

    # "r" means to reverse terms.
    if mnem.endswith(("r", "rp")):
        term1, term2 = term2, term1

    # Compute.
    try:
        result = op(term1, term2)
    except OverflowError:
        logger.debug("OVERFLOW Detected. TODO")
        result = cpu_context.registers.fpu.NaN
    except ZeroDivisionError:
        logger.debug("ZERO DIVISION detected. TODO")
        result = cpu_context.registers.fpu.NaN
    except TypeError:
        # (occurs if one of the terms was None)
        logger.debug("EMPTY value detected. TODO")
        result = cpu_context.registers.fpu.NaN

    # Store results.
    if not operands:
        cpu_context.registers.st1 = result  # storing in st1 so it's available after pop.
    elif len(operands) == 1:
        cpu_context.registers.st0 = result
    else:
        operands[0].value = result
    logger.debug("{} 0x{:X} :: {} {} {} = {}".format(mnem, ip, term1, op_str, term2, result))

    # Pop if mnem ends with "p"
    if mnem.endswith("p"):
        cpu_context.registers.fpu.pop()


@opcode
def FABS(cpu_context, ip, mnem, operands):
    """Absolute value of st(0)"""
    term = cpu_context.registers.st0
    result = abs(term)
    cpu_context.registers.st0 = result
    logger.debug("{} 0x{:X} :: abs({}) = {}".format(mnem, ip, term, result))


@opcode
def FCHS(cpu_context, ip, mnem, operands):
    """Change the sign of st(0)"""
    term = cpu_context.registers.st0
    result = -term
    cpu_context.registers.st0 = result
    logger.debug("{} 0x{:X} :: -({}) = {}".format(mnem, ip, term, result))


@opcode("fcom")
@opcode("fcomp")
@opcode("fcompp")
@opcode("fucom")
@opcode("fucomp")
@opcode("fucompp")
@opcode("fcomi")
@opcode("fcomip")
@opcode("fucomi")
@opcode("fucomip")
@opcode("ficom")
@opcode("ficomp")
@opcode("ftst")
def FCOM(cpu_context, ip, mnem, operands):
    """Compare st0 to a floating point value."""
    if not operands:
        term1 = cpu_context.registers.st0
        term2 = 0.0 if mnem == "ftst" else cpu_context.registers.st1
    elif len(operands) == 1:
        term1 = cpu_context.registers.st0
        term2 = operands[0].value
    elif len(operands) == 2:
        term1 = operands[0].value
        term2 = operands[1].value
    else:
        logger.debug("{} 0x{:X} :: Unexpected number of operands: {}".format(mnem, ip, len(operands)))
        return

    # TODO: If either value is empty (ie. None) we must set C3, C2, and C0 to None
    invalid = (None, cpu_context.registers.fpu.NaN)

    if "comi" in mnem:
        flags = ["zf", "pf", "cf"]
    else:
        flags = ["c3", "c2", "c0"]

    if term1 in invalid or term2 in invalid:
        cpu_context.registers[flags[0]] = 1
        cpu_context.registers[flags[1]] = 1
        cpu_context.registers[flags[2]] = 1
    elif term1 > term2:
        cpu_context.registers[flags[0]] = 0
        cpu_context.registers[flags[1]] = 0
        cpu_context.registers[flags[2]] = 0
    elif term1 < term2:
        cpu_context.registers[flags[0]] = 0
        cpu_context.registers[flags[1]] = 0
        cpu_context.registers[flags[2]] = 1
    elif term1 == term2:
        cpu_context.registers[flags[0]] = 1
        cpu_context.registers[flags[1]] = 0
        cpu_context.registers[flags[2]] = 0

    # Pop off st0.
    if mnem.endswith("p"):
        cpu_context.registers.fpu.pop()
    # Pop off st1 as well.
    if mnem.endswith("pp"):
        cpu_context.registers.fpu.pop()

    logger.debug("{} 0x{:X} :: Comparing: {} <-> {}".format(mnem, ip, term1, term2))


@opcode("fcmovb")
@opcode("fcmove")
@opcode("fcmovbe")
@opcode("fcmovu")
@opcode("fcmovnb")
@opcode("fcmovne")
@opcode("fcmovnbe")
@opcode("fcmovnu")
def FCMOV(cpu_context, ip, mnem, operands):
    """Conditional move based on CPU flags."""
    condition_str = mnem[5:]
    condition = False

    if "b" in condition_str:
        condition |= cpu_context.registers.cf == 1

    if "e" in condition_str:
        condition |= cpu_context.registers.zf == 1

    if "u" in condition_str:
        condition |= cpu_context.registers.pf == 1

    if condition_str.startswith("n"):
        condition = not condition

    value = operands[1].value
    if condition:
        operands[0].value = value
        logger.debug("{} 0x{:X} :: Moving: {} -> st0".format(mnem, ip, value))
    else:
        logger.debug("{} 0x{:X} :: Not moving: {} -> st0. Condition failed.".format(mnem, ip, value))


@opcode("fld")
@opcode("fild")
@opcode("fldz")
@opcode("fld1")
@opcode("fldpi")
@opcode("fldl2e")
@opcode("fldl2t")
@opcode("fldlg2")
@opcode("fldln2")
def FLD(cpu_context, ip, mnem, operands):
    """Load (push) real or integer number into stack."""
    value = orig_value = operands[0].value if operands else None
    if mnem == "fld":
        value = utils.int_to_float(value)
    elif mnem == "fild":
        value = float(value)
    elif mnem.endswith("z"):
        value = 0.0
    elif mnem.endswith("1"):
        value = 1.0
    elif mnem.endswith("pi"):
        value = 3.141592653589793  # math.pi
    elif mnem.endswith("l2e"):
        value = 1.4426950408889634  # math.log(math.e, 2)
    elif mnem.endswith("l2t"):
        value = 3.3219280948873626  # math.log(10, 2)
    elif mnem.endswith("lg2"):
        value = 0.30102999566398114  # math.log(2, 10)
    elif mnem.endswith("ln2"):
        value = 0.6931471805599453  # math.log(2) = ln(2)
    else:
        raise NotImplementedError("Unsupported mnem: {}".format(mnem))
    logger.debug("{} 0x{:X} :: Loading: {} -> {} -> st0".format(mnem, ip, orig_value, value))
    cpu_context.registers.fpu.push(value)


@opcode
def FLDCW(cpu_context, ip, mnem, operands):
    """Load control word from memory."""
    value = operands[0].value
    cpu_context.registers.fpu.control_word = value
    logger.debug("{} 0x{:X} :: Load control word: {}".format(mnem, ip, value))


@opcode("fst")
@opcode("fstp")
@opcode("fist")
@opcode("fistp")
def FST(cpu_context, ip, mnem, operands):
    """Store (pop) real or integer number from stack into into memory"""
    value = orig_value = cpu_context.registers.st0
    if "i" in mnem:
        # Round integer.
        # TODO: Technically we are suppose to round the number according to the rounding mode of rc.
        value = int(value)
    else:
        value = utils.float_to_int(value)
    operands[0].value = value
    logger.debug("{} 0x{:X} :: Storing: {} -> {} -> {}".format(mnem, ip, orig_value, value, operands[0].text))
    if mnem.endswith("p"):
        cpu_context.registers.fpu.pop()


@opcode("fstcw")
@opcode("fnstcw")
def FSTCW(cpu_context, ip, mnem, operands):
    """Store control word into memory."""
    value = cpu_context.registers.fpu.control_word
    operands[0].value = value
    logger.debug("{} 0x{:X} :: Store control word: {} -> {}".format(mnem, ip, value, operands[0].text))


@opcode
def FXAM(cpu_context, ip, mnem, operands):
    """Examine the content of st0."""
    st0 = cpu_context.registers.st0

    cpu_context.registers.c1 = int(st0 < 0)  # sign bit
    if st0 == cpu_context.registers.fpu.NaN:
        cpu_context.registers.c3 = 0
        cpu_context.registers.c2 = 0
        cpu_context.registers.c0 = 1
    elif st0 == cpu_context.registers.fpu.INFINITY:
        cpu_context.registers.c3 = 0
        cpu_context.registers.c2 = 1
        cpu_context.registers.c0 = 1
    elif st0 == 0:
        cpu_context.registers.c3 = 1
        cpu_context.registers.c2 = 0
        cpu_context.registers.c0 = 0
    elif st0 is None:
        cpu_context.registers.c3 = 1
        cpu_context.registers.c2 = 0
        cpu_context.registers.c0 = 1
    # TODO
    # elif st0 is denormalized:
    #     cpu_context.registers.c3 = 1
    #     cpu_context.registers.c2 = 1
    #     cpu_context.registers.c0 = 0
    else:
        cpu_context.registers.c3 = 0
        cpu_context.registers.c2 = 0
        cpu_context.registers.c0 = 0

    logger.debug("{} 0x{:X} :: Examining: {}".format(mnem, ip, st0))


# TODO: This is suppose to exception if st0 is empty.
@opcode
def FXCH(cpu_context, ip, mnem, operands):
    """Exchange the top data register with another data register"""
    st0 = cpu_context.registers.st0

    if operands:
        opvalue = operands[0].value
        cpu_context.registers.st0, operands[0].value = opvalue, st0
        logger.debug("{} 0x{:X} :: exchange {} <-> {}".format(mnem, ip, st0, opvalue))
    else:
        st1 = cpu_context.registers.st1
        cpu_context.registers.st0, cpu_context.registers.st1 = st1, st0
        logger.debug("{} 0x{:X} :: exchange {} <-> {}".format(mnem, ip, st0, st1))


@opcode
def SAHF(cpu_context, ip, mnem, operands):
    """Transfer status word flags into CPU's flag register."""
    cpu_context.registers.zf = cpu_context.registers.c3
    cpu_context.registers.pf = cpu_context.registers.c2
    cpu_context.registers.cf = cpu_context.registers.c0



"""
Interface for variable management.
"""

from builtins import map
from builtins import range
from builtins import object
from copy import deepcopy
import functools
import logging

import idc

from kordesii.utils.function_tracing import utils


logger = logging.getLogger(__name__)


class VariableMap(object):
    """Class that stores a set of variables that have been encountered during emulation."""

    def __init__(self, cpu_context):
        self._variables = {}
        self._cpu_context = cpu_context

    def __repr__(self):
        return "<VariableMap : \n\t{}\n>".format(
            "\n\t".join(map(repr, sorted(list(self._variables.values()), key=lambda var: var.addr)))
        )

    def __deepcopy__(self, memo):
        """
        Custom implementation of deepcopy to improve efficiency.
        """
        copy = VariableMap(deepcopy(self._cpu_context, memo))
        memo[id(self)] = copy
        copy._variables = {addr: deepcopy(variable, memo) for addr, variable in list(self._variables.items())}
        return copy

    def __getitem__(self, addr_or_name):
        """Gets a variable by name or address."""
        if isinstance(addr_or_name, str):
            for var in list(self._variables.values()):
                if addr_or_name == var.name:
                    return var
            raise KeyError("{} not found.".format(addr_or_name))
        elif isinstance(addr_or_name, int):
            return self._variables[addr_or_name]
        else:
            raise ValueError("Invalid variable name or address: {!r}".format(addr_or_name))

    def get(self, addr_or_name, default=None):
        """Gets a variable by name or address."""
        try:
            return self[addr_or_name]
        except (KeyError, ValueError):
            return default

    def at(self, ip):
        """
        Retrieves the variables referenced at the given instruction address.

        :param ip: Instruction address to get pointers from.
        :return: List of Variable objects that were found within the given instruction.

        :raises ValueError: If instruction has not been executed yet.
        """
        if ip not in self._cpu_context.executed_instructions:
            raise ValueError("Unable to get variables. Instruction at 0x{:0x} has not been executed.".format(ip))

        return [var for var in self if ip in var.references]

    def __setitem__(self, addr_or_name, variable):
        """Sets a variable by name or address."""
        if isinstance(addr_or_name, str):  # TODO
            raise NotImplementedError("Creating new variable by name is currently not supported.")
        elif isinstance(addr_or_name, int):
            self._variables[addr_or_name] = variable
        else:
            raise ValueError("Invalid variable name or address: {!r}".format(addr_or_name))

    def __contains__(self, addr_or_name):
        if isinstance(addr_or_name, str):
            for var in list(self._variables.values()):
                if addr_or_name == var.name:
                    return True
            return False
        elif isinstance(addr_or_name, int):
            return addr_or_name in self._variables
        else:
            raise ValueError("Invalid variable name or address: {!r}".format(addr_or_name))

    def __iter__(self):
        return iter(list(self._variables.values()))

    def add(self, addr, frame_id=None, stack_offset=None, reference=None):
        """
        Creates and adds a variable object to mapping by object

        If the variable already exists, this function does nothing.

        :return: Variable object that has been created or one that already exists.
        """
        if addr in self._variables:
            var = self._variables[addr]
        else:
            var = Variable(self._cpu_context, addr, frame_id=frame_id, stack_offset=stack_offset)
            # logger.debug('VariableMap :: Created variable: {!r}'.format(var))
            self._variables[addr] = var
        if reference:
            var.add_reference(reference)
        return var

    @property
    def names(self):
        return [var.name for var in list(self._variables.values())]

    @property
    def addrs(self):
        return list(self._variables.keys())

    @property
    def stack_variables(self):
        return [var for var in list(self._variables.values()) if var.is_stack]

    @property
    def global_variables(self):
        return [var for var in list(self._variables.values()) if not var.is_stack]


@functools.total_ordering
class Variable(object):
    """Stores information for a local / global variable for a specific CPU context state."""

    # Maps data types to their sizes.
    SIZE_MAP = {
        idc.FF_BYTE: 1,
        idc.FF_WORD: 2,
        idc.FF_DWORD: 4,
        idc.FF_QWORD: 8,
        idc.FF_OWORD: 16,
    }

    def __init__(self, cpu_context, addr, frame_id=None, stack_offset=None):
        if (frame_id is not None and stack_offset is None) or (frame_id is None and stack_offset is not None):
            raise ValueError("Both frame_id and stack_offset must be provided.")
        self._cpu_context = cpu_context
        self.addr = addr
        self.frame_id = frame_id
        self.stack_offset = stack_offset
        self.references = []  # list of instruction pointers where the variable was encountered.

    def __deepcopy__(self, memo):
        copy = self.__new__(self.__class__)
        memo[id(self)] = copy
        copy._cpu_context = deepcopy(self._cpu_context, memo)
        copy.addr = self.addr
        copy.frame_id = self.frame_id
        copy.stack_offset = self.stack_offset
        copy.references = list(self.references)
        return copy

    def __repr__(self):
        string = "<Variable {} : addr = 0x{:0x} : value = {!r} : size = {}".format(
            self.name, self.addr, self.value, self.size
        )
        if self.is_stack:
            string += " : frame_id = 0x{:0x} : stack_offset = {}".format(self.frame_id, self.stack_offset)
        string += ">"
        return string

    def __eq__(self, other):
        return self.addr == other.addr

    def __lt__(self, other):
        return self.addr < other.addr

    @property
    def is_stack(self):
        """True if variable is on stack."""
        return self.stack_offset is not None

    @property
    def is_func_ptr(self):
        return utils.is_func_ptr(self.addr)

    @property
    def name(self):
        if self.is_stack:
            name = idc.get_member_name(self.frame_id, self.stack_offset)
        else:
            name = idc.get_name(self.addr)
        if not name:
            return ""
        return name

    @property
    def size(self):
        if self.is_stack:
            return idc.get_member_size(self.frame_id, self.stack_offset)
        else:
            return idc.get_item_size(self.addr)

    @property
    def data(self):
        """The raw data the variable is pointing to."""
        return self._cpu_context.mem_read(self.addr, self.size)

    def add_reference(self, ip):
        """Adds ip to list of references for this variable."""
        # Ignore duplicate calls.
        if self.references and ip == self.references[-1]:
            return
        self.references.append(ip)

    @property
    def value(self):
        """The unpacked data the variable is pointing to."""
        if self.is_func_ptr:
            return self.addr

        if self.is_stack:
            flag = idc.get_member_flag(self.frame_id, self.stack_offset)
            data_type = flag & idc.DT_TYPE
            # Unpack if an integer type.
            if data_type in self.SIZE_MAP:
                data_type_size = self.SIZE_MAP[data_type]
                if self.size == data_type_size:
                    return utils.struct_unpack(self.data)
                else:
                    # If data size is greater than type size, then we have an array.
                    data = self.data
                    return [
                        utils.struct_unpack(data[i : i + data_type_size]) for i in range(0, len(data), data_type_size)
                    ]
            else:
                return self.data
        else:
            # TODO: Determine how to unpack based on type for global variables.
            return self.data

    @property
    def history(self):
        """The history of variables by following memory copies."""
        # (We shouldn't have Nones)
        history = []
        for addr in self._cpu_context.get_pointer_history(self.addr):
            var = self._cpu_context.variables.get(addr, None)
            if var:
                history.append(var)
        return history



"""
Experimental new stack string decoder using function_tracing
"""

import io

import idc

import kordesii
from kordesii.utils import decoderutils
from kordesii.utils import function_tracing
from kordesii.utils import utils


logger = kordesii.get_logger()


ENCODINGS = [("utf-8", 1), ("utf-16-le", 2)]


def num_raw_bytes(string):
    """
    Returns the number of raw bytes found in the given unicode string
    """
    count = 0
    for char in string:
        char = char.encode("unicode-escape")
        count += char.startswith(b"\\x") + char.startswith(b"\\u") * 2
    return count


def read_string(data):
    """
    Read data until we find a something that is not a printable ascii character.

    :return: String and encoding if we find a string of at least 1 character.
        Returns Nones otherwise.
    """
    stream = io.BytesIO(data)
    strings = []
    for encoding, width in ENCODINGS:
        stream.seek(0)
        string = u""
        while True:
            char = stream.read(width)
            if not char:
                # Ran out of bytes
                break
            try:
                char = char.decode(encoding)
            except UnicodeDecodeError:
                break
            if char == u"\0" or num_raw_bytes(char):
                break
            string += char
        if string:
            strings.append((string, encoding))

    if not strings:
        return None, None

    # Return whichever encoding uses the most data.
    return max(strings, key=lambda s: len(s[0]))


class StackStringExtractor(object):
    def __init__(self):
        self.encoded_strings = []

    def process_string(self, context, var, ip):
        # Read in data from stack and see if it's a valid string.
        stack_data = context.read_data(var.addr, 1024)
        string, encoding = read_string(stack_data)
        if string:
            data = string.encode(encoding)
            encoded_string = decoderutils.EncodedStackString(
                data, frame_id=var.frame_id, stack_offset=var.stack_offset, string_reference=ip, code_page=encoding
            )
            encoded_string.decoded_data = data
            self.encoded_strings.append((var.addr, encoded_string))

    def parse_stack_strings(self, func):

        logger.debug("Processing function: 0x{:X}".format(func.start_ea))
        tracer = function_tracing.get_tracer(func.start_ea)

        waiting_for_call = []

        context = None
        for ea in func.heads():
            context = tracer.context_at(ea)
            if not context:
                continue
            context.execute()  # also include instruction we are looking at.

            # If we encounter a call, process pushed in variables.
            if idc.print_insn_mnem(ea) == "call":
                for ip, var in waiting_for_call:
                    self.process_string(context, var, ip)
                waiting_for_call = []
                continue

            # Look for instructions where a stack variable is being used for something other than
            # a move.
            # We can do this by only considering variables that are the last operand.
            operands = context.get_operands(ea)
            if not operands:
                continue
            addr = operands[-1].addr or operands[-1].value
            if not addr:
                continue
            var = context.variables.get(addr)
            if var and var.is_stack:
                # Ignore string if it comes from memory with no concatinations.
                if var.history and idc.is_loaded(var.history[0].addr):
                    continue

                # If instruction is a push, it is possible that the string will be populated
                # after this instruction. Therefore, wait for the function call be before processing.
                if idc.print_insn_mnem(ea) == "push":
                    waiting_for_call.append((ea, var))
                else:
                    self.process_string(context, var, ea)

        # Process any strings still waiting for a call.
        if context:
            for ip, var in waiting_for_call:
                self.process_string(context, var, ip)

        # Remove any substrings or strings that are too small.
        for addr, encoded_string in sorted(self.encoded_strings):
            if len(encoded_string.encoded_data) < 3:
                self.encoded_strings.remove((addr, encoded_string))
                continue
            for _addr, _encoded_string in self.encoded_strings[:]:
                # Remove dups
                if (
                    _addr == addr
                    and _encoded_string is not encoded_string
                    and _encoded_string.encoded_data == encoded_string.encoded_data
                ):
                    self.encoded_strings.remove((addr, encoded_string))
                    break
                # Remove substrings
                elif _addr < addr:
                    index = addr - _addr
                    substring = _encoded_string.encoded_data[index : index + len(encoded_string.encoded_data)]
                    if substring == encoded_string.encoded_data:
                        self.encoded_strings.remove((addr, encoded_string))
                        break

        # Report found strings
        for _, encoded_string in sorted(self.encoded_strings):
            # Don't want to rename because the buffers could be reused for multiple strings.
            encoded_string.publish(rename=False, patch=False)
            # TODO: EncodedString should allow commenting without renaming.
            idc.set_cmt(encoded_string.string_reference, 'Stack String: "{}"'.format(encoded_string.display_name), 0)


@kordesii.decoder_entry
def main():

    # NOP memset, its emulation is not needed and it's slowing things down.
    def memset(context, func_name, func_args):
        return

    function_tracing.hook_tracers("memset", memset)

    for ea, name in utils.iter_functions():
        try:
            func = decoderutils.SuperFunc_t(ea)
        except AttributeError:
            continue
        if not func.is_library:
            StackStringExtractor().parse_stack_strings(func)



import abc
import collections
import copy
import functools
import hashlib
import logging
import os
import re
import sys

import ida_bytes
import ida_name
import idaapi
import idautils
import idc

import kordesii
from kordesii.serialization import serializable_class
from kordesii.utils import function_tracing
from kordesii.utils import yara
from kordesii.utils.function_creator import create_function_precise
from kordesii.utils import ida_re
from kordesii.serialization import serializable_class

logger = logging.getLogger(__name__)

# fmt: off
# Codecs used to detect encoding of strings.
CODE_PAGES = [
    "ascii",
    "utf-32-be", "utf-32-le", "utf-16-be", "utf-16-le", "utf-8",  # General (utf-7 omitted)
    "gb18030", "gbk",  # Unified Chinese
    "gb2312", "hz",  # Simplified Chinese
    "big5hkscs", "big5",  # Traditional Chinese (cp950 omitted)
    "koi8-r", "iso8859-5", "cp1251", "mac-cyrillic",  # Cyrillic (cp866, cp855 omitted)
    "cp949",  # Korean (johab, iso2022-kr omitted)
    "iso8859-6", "cp1256",  # Arabic (cp864, cp720 omitted)
    "latin1",  # If all else fails, latin1 is always is successful.
]
# fmt: on

INPUT_FILE_PATH = idc.get_input_file_path()
# Put these here for increased robustness. Please don't depend on these very often.
ENCODED_STRINGS = []
DECODED_STRINGS = []

# CODEC to use for displaying strings in IDA, etc.
DISPLAY_CODE = "cp437" if sys.platform == "win32" else "ascii"


class SuperFunc_t(object):
    """
    Description:
        Effectively extends func_t to also know its name and all its non-recursive xrefs and knows
        how to rename itself.

    Fields:
        function_obj - The idaapi.func_t object for this function. (<ea> must be within a function.)
        name - The name of the function.
        xrefs_to - EA's for all of the non-recursive references to this function's start_ea.
        xref_count - len(xrefs_to)

    Input:
        ea - An EA within the function.
        identifier - The id of the YARA rule that hit in this function
        create_if_not_exists - If true, uses IN_Dev_Repo's function creator to create a function containing <ea>
    """

    def __init__(self, ea, identifier=None, create_if_not_exists=True):
        super(SuperFunc_t, self).__init__()
        self.origin_ea = ea
        self.identifier = identifier
        self.function_obj = idaapi.get_func(ea)
        if not self.function_obj:
            if create_if_not_exists:
                if create_function_precise(ea, False):
                    self.function_obj = idaapi.get_func(ea)
                    logger.debug(
                        "Created function at 0x%X" % self.function_obj.start_ea
                    )
                else:
                    raise AttributeError("No function at 0x%X" % ea)
            else:
                raise AttributeError("No function at 0x%X" % ea)
        if self.function_obj:
            self.start_ea = self.function_obj.start_ea
            self.end_ea = self.function_obj.end_ea
        self.name = idaapi.get_func_name(self.function_obj.start_ea)
        self.xrefs_to = [
            ref.frm
            for ref in idautils.XrefsTo(self.function_obj.start_ea)
            if idaapi.get_func_name(ref.frm) != self.name
        ]
        self.xref_count = len(self.xrefs_to)
        self._flowchart = None
        self._api_calls = None

    @classmethod
    def from_name(cls, func_name, ignore_underscore=False):
        """
        Factory method for obtaining SuperFunc_t by name.

        :param str func_name: Name of function to obtain
        :param bol ignore_underscore: Whether to ignore underscores in function name.
            (Will return the first found function if enabled.)

        :return: SuperFunc_t object
        :raises ValueError: If function name was not found.
        """
        for ea in idautils.Functions():
            _func_name = idc.get_func_name(ea)
            if ignore_underscore:
                _func_name = _func_name.strip("_")
            if func_name == _func_name:
                return cls(ea)
        raise ValueError("Unable to find function with name: {}".format(func_name))

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return self.__repr__().__hash__()

    def __str__(self):
        return "%s 0x%X - 0x%X" % (
            self.name,
            self.function_obj.start_ea,
            self.function_obj.end_ea,
        )

    def __repr__(self):
        return "<SuperFunc_t : {}() : {:#08x} - {:#08x}>".format(
            self.name, self.start_ea, self.end_ea
        )

    def __contains__(self, ea):
        """Tests if ea is within function."""
        # NOTE: We can't just test if it's between start_ea and end_ea because the function might have
        # fragmented function chunks.
        try:
            func = idaapi.get_func(ea)
            if not func:
                return False
        except AttributeError:
            return False
        return func.start_ea == self.start_ea

    def heads(self, start=None, reverse=False):
        """
        Iterates all the heads for the given function.

        :param start: Start address (defaults to start_ea or end_ea)
        :param reverse:  Direction to iterate
        :return:
        """
        if not self._flowchart:
            self._flowchart = function_tracing.FlowChart(self.start_ea)
        if not start:
            start = self.end_ea if reverse else self.start_ea

        for ea in self._flowchart.heads(start, reverse=reverse):
            yield ea

    def rename(self, new_name):
        """
        Attempts to apply new_name to the object at <ea>. If more than one object starts at <ea>, the
        largest object will be renamed. If that name already exists, let IDA resolve the collision
        and then return that name. If new_name is "", reset the name to IDA's default.

        :param str new_name: The desired new name for the function.

        :return str: The name that ended up getting set (unless no name was set, then return None).
        """
        if new_name == "":
            if idaapi.set_name(self.start_ea, new_name):
                return idaapi.get_name(self.function_obj.start_ea)
            else:
                logger.warning("Failed to reset name at 0x%X" % self.start_ea)
        elif ida_name.force_name(self.start_ea, new_name):
            self.name = idaapi.get_name(self.start_ea)
            if self.name != new_name:
                logger.info('IDA changed name "%s" to "%s"' % (new_name, self.name))
            return self.name
        else:
            logger.warning("Failed to rename at 0x%X" % self.start_ea)

    @property
    def api_calls(self):
        """
        Returns counter containing API calls and the number of times they were called.
        """
        if self._api_calls:
            return self._api_calls

        api_calls = collections.Counter()
        for ea in self.heads():
            if idc.print_insn_mnem(ea) == "call":
                for xref in idautils.XrefsFrom(ea, idaapi.XREF_FAR):
                    if xref.to:
                        func_name = ida_name.get_name(xref.to)
                        if func_name:
                            api_calls.update([func_name])

        self._api_calls = api_calls
        return self._api_calls

    @property
    def is_library(self):
        """
        Is the function a library?
        """
        return bool(self.function_obj.flags & idc.FUNC_LIB)

    @property
    def calls_to(self):
        """Iterates addresses that call this function."""
        for ea in self.xrefs_to:
            if idc.print_insn_mnem(ea) == "call":
                yield ea

    @property
    def callers(self):
        """Iterates SuperFunc_t objects that call this function."""
        cache = set()
        for ea in self.calls_to:
            try:
                func = SuperFunc_t(ea)
            except AttributeError:
                continue
            if func.name not in cache:
                yield func
                cache.add(func.name)

    @property
    def data(self):
        """Returns all the bytes contained in the function."""
        return ida_bytes.get_bytes(self.start_ea, self.end_ea - self.start_ea)


@serializable_class
@functools.total_ordering
class EncodedString(object):
    """
    Description:
        Object to hold data about an encoded/encrypted string.

    Fields:
        string_location - The EA at which the encoded_data starts.
        string_reference - The EA from which the string is referenced. Defaults to None.
        size - The size of the encoded_data. When set in the constructor, causes encoded_data to be
               populated. Defaults to None.
        offset - The offset from string_location at which the actual encoded_data starts. Often unused.
                 Defaults to None.
        key - The key used for decoding. Generally, this os only set when the key differs by string.
        encoded_data - The raw data from the file that we intend to decode.
            Automatically populated from string_location if not manually provided.
        decoded_data - The string's value after it has been decoded/decrypted.
        code_page - Code page used to decode string when unicode() is called.

    Input:
        string_location - The location of the encrypted data.
            Usually required, but can be None if you are providing an encoded_data.
        string_reference - The location the string is referenced from. Often helpful
        size - Known size of the encoded data. Used to retrieve data if encoded_data is not provided.
        offset - Used when there is an offset based accessing scheme.
        key - Used when there is a key that can vary by string.
        encoded_data - encoded/encrypted data, if not provided data will be retrieved from IDA.
        code_page - known encoding page used to decode data to unicode (after data is decrypted)
            (code page is dynamically determined if not provided)
        dest - Location of decrypted data (if different from string_location)

    Raises:
        RuntimeError - If encoded data cannot be extracted.
    """

    _MAX_COMMENT_LENGTH = 130
    _MAX_NAME_LENGTH = 30

    # TODO: Since we must have either a string_location or encoded_data should we change
    #   the first parameter to be something like "encoded_data_or_location" and then
    #   check which it is by type?
    def __init__(
        self,
        string_location,
        string_reference=None,
        size=None,
        offset=None,
        key=None,
        encoded_data=None,
        code_page=None,
        dest=None,
    ):
        self.string_location = string_location
        self.string_reference = string_reference
        self.offset = offset
        self.key = key
        self.decoded_data = None
        self.code_page = None
        self.dest = dest
        self._xrefs_to = None

        # Pull encoded_data from IDA if not provided.
        if encoded_data is None:
            if string_location is None:
                raise ValueError(
                    "encoded_data must be provided if string_location is None"
                )
            encoded_data = (
                self._get_bytes(string_location, size=size, code_page=code_page) or b""
            )

        if not isinstance(encoded_data, (bytes, list)):
            raise TypeError(
                "encoded_data must be type 'bytes', not '{!r}'".format(
                    type(encoded_data)
                )
            )

        self.encoded_data = encoded_data

    @classmethod
    def factory(
        cls,
        string_location,
        string_reference,
        size=None,
        offset=None,
        key=None,
        encoded_data=None,
        code_page=None,
        dest=None,
    ):
        """
        Factory function to generate an EncodedString or EncodedStackString based on type.

        :param string_location:
            Data segment pointer for static strings or stack pointer for stack strings.
        :param string_reference:
            The location the string is referenced from.
            This is required to pull the stack frame when string_location is a stack pointer.
        :param size: The size of the string. Required to use self.get_bytes.
        :param offset: Used when there is an offset based accessing scheme.
        :param key: Used when there is a key that can vary by string.
        :param encoded_data: encoded/encrypted data, if not provided data will be retrieved from IDA.
        :param code_page: known encoding page used to decode data to unicode (after data is decrypted)
            (code page is dynamically determined if not provided)
        :param dest: Location of decrypted data (if different from string_location)
        """
        if idc.is_loaded(string_location):
            return EncodedString(
                string_location,
                string_reference,
                size=size,
                offset=offset,
                key=key,
                encoded_data=encoded_data,
                code_page=code_page,
                dest=dest,
            )

        # otherwise assume string_location is a pointer within the stack
        # (using function_tracing's CPU emulator) and create an EncodedStackString object.
        stack = idc.get_func_attr(string_reference, idc.FUNCATTR_FRAME)
        # FIXME: This method isn't always super accurate because... IDA
        stack_offset = (
            string_location
            + function_tracing.RSP_OFFSET
            + idc.get_func_attr(string_reference, idc.FUNCATTR_FRSIZE)
            - function_tracing.STACK_BASE
        )
        if stack_offset < 0:
            logger.warning(
                "Ignoring negative stack offset {:#x} pulled from 0x{:X}".format(
                    stack_offset, string_location
                )
            )
            stack_offset = None

        return EncodedStackString(
            encoded_data,
            frame_id=stack,
            stack_offset=stack_offset,
            string_reference=string_reference,
            size=size,
            offset=offset,
            key=key,
            code_page=code_page,
            dest=dest,
        )

    def _compare_key(self):
        # Sort by where it was found then where it is referenced
        return (
            self.string_location or -1,
            self.string_reference or -1,
            self.decoded_data or b"",
        )

    def __hash__(self):
        return hash(self._compare_key())

    def __eq__(self, other):
        return self._compare_key() == other._compare_key()

    def __lt__(self, other):
        return self._compare_key() < other._compare_key()

    def report(self):
        """
        Generates a text report of the EncodedString object.

        :return str: Unicode string containing text report.
        """
        # We have to return repr as a unicode because of the decoded string.
        text = u""
        if self.string_location is not None:
            text += u"EA:  0x{:08X}\n".format(self.string_location)
        if self.string_reference is not None:
            text += u"Ref: 0x{:08X}\n".format(self.string_reference)
        if self.dest is not None:
            text += u"Dest: 0x{:08X}\n".format(self.dest)
        if self.offset is not None:
            text += u"Offset: 0x{:08X}\n".format(self.offset)
        if self.encoded_data:
            text += u"Raw Enc: {!r}\n".format(self.encoded_data)
        if self.decoded_data:
            text += u"Raw Dec: {!r}\n".format(self.decoded_data)
            dec_string = str(self)
            if self.code_page:
                text += u"Detected Code Page: {}\n".format(self.code_page)
            text += u"Dec: {}".format(dec_string)
        return text

    def __bytes__(self):
        return self.decoded_data

    def __str__(self):
        return self._decode_unknown_charset()

    def __unicode__(self):
        return self._decode_unknown_charset()

    def __repr__(self):
        encoded_data = self.encoded_data or b""
        if len(encoded_data) > 30:
            encoded_data = encoded_data[:30] + b" ..."
        return "<{!r} at 0x{:08x}>".format(
            encoded_data, self.string_location or self.string_reference
        )

    @property
    def xrefs_to(self):
        if self._xrefs_to is None:
            self._xrefs_to = [ref.frm for ref in idautils.XrefsTo(self.string_location)]
        return self._xrefs_to

    @property
    def display_name(self):
        """Returns an IDA friendly, printable name for the decoded string."""
        return str(self).encode(DISPLAY_CODE, "replace").decode(DISPLAY_CODE)

    def rename(self, name=None):
        """
        Renames (and comments) the string variable in IDA.

        :param str name: New name to given encoded string. (defaults to decoded_string)
        """
        name = name or self.display_name
        if not name:
            logger.warning(
                "Unable to rename encoded string due to no decoded string: {!r}".format(
                    self
                )
            )
            return

        # Add comment
        comment = '"{}"'.format(name[: self._MAX_COMMENT_LENGTH])
        if len(name) > self._MAX_COMMENT_LENGTH:
            comment += " (truncated)"
        if self.string_location is not None:
            idc.set_cmt(self.string_location, comment, 1)
        if self.string_reference is not None:
            idc.set_cmt(self.string_reference, comment, 1)
        if self.dest is not None:
            idc.set_cmt(self.dest, comment, 1)

        # Set variable name
        # To follow IDA conventions, the decoded string itself will be prefixed with 'a'
        # If a destination was provided, don't include any prefix, because this
        # is usually for resolved API functions.
        name = name[: self._MAX_NAME_LENGTH]
        if self.string_location is not None:
            ida_name.force_name(self.string_location, "a" + name)
        if self.dest is not None:
            ida_name.force_name(self.dest, name)

    def patch(self, fill_char=None, define=True):
        """
        Patches the original encoded string with the decoded string.

        :param str fill_char:
            Character to use to fill left over space if decoded data
            is shorter than its encoded data. (defaults to leaving the original data)
        :param bool define: Whether to define the string after patching.
        """
        if not self.decoded_data or self.string_location is None:
            return
        decoded_data = self.decoded_data
        if fill_char:
            decoded_data += fill_char * (len(self.encoded_data) - len(decoded_data))
        try:
            idaapi.patch_bytes(self.start_ea, decoded_data)
            if define:
                self.define()
        except TypeError:
            logger.debug(
                "String type for decoded string from location 0x{:08x}.".format(
                    self.start_ea
                )
            )
        finally:
            return self

    # TODO: Can this just be embedded in patch()?
    def define(self):
        """
        Defines the string in the IDB.
        """
        try:
            idc.del_items(self.start_ea, idc.DELIT_SIMPLE, len(self.decoded_data))
            idaapi.create_strlit(
                self.start_ea, len(self.decoded_data), self.string_type
            )
        except Exception as e:
            logger.warning(
                "Unable to define string at 0x{:0X}: {}".format(self.start_ea, e)
            )

    def publish(self, rename=True, patch=True):
        """
        - Saves encoded string to external kordesii Reporter
        - Prints a report about the string to the console
        - renames and patches the IDB with decoded data

        :param rename: Whether to rename the string in the IDB.
        :param patch: Whether to patch the string with the decoded variant in the IDB.
        """
        if not self.decoded_data:
            logger.warning(
                "Unable to publish string {!r}. Missing decoded_data.".format(self)
            )
            return

        # FIXME: Even though we strip nulls in __unicode__(), there still seems to be some strings
        # with null characters seeping through.
        kordesii.append_string(self._decode_unknown_charset().rstrip(u"\x00"))

        print("\n")
        display = self.report()
        print(display)

        if rename:
            self.rename()
        if patch:
            self.patch()

    @property
    def start_ea(self):
        if self.string_location in (None, idc.BADADDR):
            return -1
        else:
            return self.string_location + (self.offset or 0)

    @property
    def end_ea(self):
        start_ea = self.start_ea
        if start_ea is None:
            return -1
        else:
            return start_ea + len(self.decoded_data)

    @property
    def string_type(self):
        if not self.decoded_data:
            return None
        else:
            return (
                idc.STRTYPE_C_16
                if isinstance(self.decoded_data, str)
                else idc.STRTYPE_C
            )

    def _get_bytes(self, location, size=None, code_page=None):
        """
        Extracts bytes from given location.

        :param location: Location to pull bytes
        :param size: Number of bytes to pull (determines size by looking for terminator if not provided)
        :param code_page: Known code_page used to determine terminator.
        :return: bytes or None
        """
        # Determine size by looking for terminator.
        # (Use provided encoding to determine terminator width.)
        if size is None:
            width = 1
            if code_page:
                if "16" in code_page:
                    width = 2
                elif "32" in code_page:
                    width = 4
            end_location = idc.find_binary(location, idc.SEARCH_DOWN, "00 " * width)
            if end_location == idc.BADADDR:
                logger.warning("Failed to extract bytes from 0x{:08X}".format(location))
                return None

            size = end_location - location
            while size % width:  # ensure unicode strings are a valid length
                size += 1

        # Pull size amount of bytes from IDA.
        data = idc.get_bytes(location, size)
        if data is None:
            logger.warning(
                "Failed to extract {} bytes from 0x{:08X}".format(size, location)
            )
        return data

    def _num_raw_bytes(self, string):
        """
        Returns the number of raw bytes found in the given unicode string
        """
        count = 0
        for char in string:
            char = char.encode("unicode-escape")
            count += char.startswith(b"\\x") + char.startswith(b"\\u") * 2
        return count

    def _decode_unknown_charset(self):
        """
        Returns a decoded string using the best guess codec.
        """
        if not self.decoded_data:
            return u""

        # First see if the decoder already gave us unicode.
        if isinstance(self.decoded_data, str):
            return self.decoded_data

        # If a code page was set (either by us or the decoder) use that.
        # If the code page doesn't work... move on.
        if self.code_page:
            try:
                return self.decoded_data.decode(self.code_page)
            except UnicodeDecodeError:
                pass

        # TODO: Use chardet if they ever support utf16 without BOM.
        best_score = len(self.decoded_data)
        best_code_page = None
        best_output = None
        for code_page in CODE_PAGES:
            try:
                output = self.decoded_data.decode(code_page).rstrip(u"\x00")
            except UnicodeDecodeError:
                # If it's UTF we may need to strip away some null characters before decoding.
                if code_page in ("utf-16-le", "utf-16-be", "utf-32-le", "utf-32-be"):
                    decoded_data = self.decoded_data
                    while decoded_data and decoded_data[-1] == 0:
                        try:
                            decoded_data = decoded_data[:-1]
                            output = decoded_data.decode(code_page).rstrip(u"\x00")
                        except UnicodeDecodeError:
                            continue
                        break  # successfully decoded
                    else:
                        continue
                # otherwise the code page isn't correct.
                else:
                    continue

            score = self._num_raw_bytes(output)
            if not best_output or score < best_score:
                best_score = score
                best_output = output
                best_code_page = code_page

        if best_output:
            self.code_page = best_code_page
            return best_output

        return u""


@serializable_class(skip_attrs=["xrefs_to"])
class EncodedStackString(EncodedString):
    """
    Variant of EncodedString that represents a string built from the stack.
    """

    def __init__(
        self,
        encoded_data,
        frame_id=None,
        stack_offset=None,
        memory_ptr=None,
        string_reference=None,
        size=None,
        offset=None,
        key=None,
        code_page=None,
        dest=None,
    ):
        super(EncodedStackString, self).__init__(
            None,
            string_reference=string_reference,
            size=size,
            offset=offset,
            key=key,
            encoded_data=encoded_data,
            code_page=code_page,
            dest=dest,
        )
        # Frame ID and Stack Offset are optional because it's not always easy to calculate this.
        self.frame_id = frame_id
        self.stack_offset = stack_offset
        # TODO: Remove memory_ptr, it should be manually added by the decoder after initialization
        # if they want to use it.
        # Optional extra argument to keep track of pointer to string in memory when using function_tracing.
        self.memory_ptr = memory_ptr

    def _compare_key(self):
        # Sort by where it was found and then by offset within stack.
        return (
            self.string_reference,
            self.frame_id,
            self.stack_offset,
            self.decoded_data,
        )

    def report(self):
        """
        General display format.

        :return unicode:
            A unicode string with the string's frame ID and stack offset,
            reference EA (where applicable), and the decoded value (where applicable).
        """
        text = u""
        if self.frame_id:
            text += u"Frame ID: 0x%X\n" % self.frame_id
        if self.stack_offset:
            text += u"Stack Offset: 0x%X\n" % self.stack_offset
        text += super(EncodedStackString, self).report()
        return text

    @property
    def xrefs_to(self):
        """
        Retrieves the xrefs to the stack variable.

        NOTE: This code is very SWIGGY because IDA did not properly expose this functionality.

        :raises ValueError: if frame_id, stack_offset, or string_reference was not provided.
            This is needed to determine what function to use.
        """
        if self._xrefs_to is None:
            if not self.string_reference:
                raise ValueError("Unable to get xrefs without string_reference.")
            if not (self.frame_id and self.stack_offset):
                raise ValueError(
                    "Unable to get xrefs without frame_id and stack_offset"
                )
            xrefs = idaapi.xreflist_t()
            frame = idaapi.get_frame(self.frame_id)
            func = idaapi.get_func(self.string_reference)
            member = idaapi.get_member(frame, self.stack_offset)
            idaapi.build_stkvar_xrefs(xrefs, func, member)
            self._xrefs_to = [ref.ea for ref in xrefs]
        return self._xrefs_to

    def rename(self, name=None):
        """
        Renames (and comments) the string variable in IDA.

        :param str name: New name to given encoded string. (defaults to decoded_string)
        """
        name = name or self.display_name
        if not name:
            logger.warning(
                "Unable to rename encoded string due to no decoded string: {!r}".format(
                    self
                )
            )

        # Set name and comment in stack variable.
        comment = '"{}"'.format(name[: self._MAX_COMMENT_LENGTH])
        if len(name) > self._MAX_COMMENT_LENGTH:
            comment += " (truncated)"
        if self.frame_id and self.stack_offset:
            idc.set_member_cmt(self.frame_id, self.stack_offset, comment, repeatable=1)
            var_name = re.sub(
                "[^_$?@0-9A-Za-z]", "_", name[: self._MAX_NAME_LENGTH]
            )  # Replace invalid characters
            if not var_name:
                raise ValueError("Unable to calculate var_name for : {!r}".format(self))
            var_name = "a" + var_name.capitalize()
            idc.set_member_name(self.frame_id, self.stack_offset, var_name)

        if self.dest is not None:
            ida_name.force_name(self.dest, name)

        # Add a comment where the string is being used.
        if self.string_reference:
            idc.set_cmt(self.string_reference, comment, 1)

    def patch(self, fill_char=None, define=True):
        """Does nothing, patching is not a thing for stack strings."""
        return self


RAX_FAM = ["rax", "eax", "ax", "ah", "al"]


def find_destination(start, instruction_limit=None):
    """
    Finds the destination address for returned eax register.

    :param int start: Starting address to start looking
    :param int instruction_limit: Limit the number of instructions to traverse before giving up.
        Defaults to searching until the end of the function.

    :return: destination address or None if address couldn't be found or is not a loaded address.
    """
    count = 0
    func = SuperFunc_t(start)
    for ea in func.heads(start):
        count += 1
        if instruction_limit is not None and count > instruction_limit:
            return None

        if idc.print_insn_mnem(ea) == "mov" and idc.print_operand(ea, 1) in RAX_FAM:
            if idc.get_operand_type(ea, 0) == idc.o_mem:
                return idc.get_operand_value(ea, 0)
            else:
                return None
    return None


class StringTracer(object, metaclass=abc.ABCMeta):
    """
    Description:
        An object to hold tracing info for one offset at a time.

        This abstract class doesn't do anything, but attempts to define fields that are
        almost always used. If a field is not used, it is recommended to set it to UNUSED.

    Fields:
        initial_offset - The EA the searching starts at. This is usually a yara match EA or a
                         func_t.start_ea. (<initial_offset> must be within a function.)
        func_ea - The start_ea of the function containing the initial_offset.
        string_location - The EA the encoded string starts at. Defaults to idc.BADADDR.
        string_reference - The EA from which the encoded string is referenced. Defaults to
                           idc.BADADDR.
        encoded_strings - Storage for the encoded string(s) found by search. Defaults to [].
        size - The size of the string. Defaults to None.

    Input:
        initial_offset - Required. We have to start somewhere.
        identifier - The id of the YARA rule that this Tracer is based off of

    Throws:
        AttributeError - There was no function at initial_offset.
    """

    def __init__(self, initial_offset=None, identifier=None):
        super(StringTracer, self).__init__()
        self.initial_offset = initial_offset
        self.identifier = identifier
        if self.initial_offset is not None:
            if not idaapi.get_func(initial_offset):
                raise AttributeError("No function at 0x%X" % initial_offset)
            self.func_ea = idaapi.get_func(initial_offset).start_ea
        else:
            self.func_ea = None
        # TODO: remove use of idc.BADADDR
        self.string_location = idc.BADADDR
        self.string_reference = idc.BADADDR
        self.encoded_strings = []
        self.size = None

    @abc.abstractmethod
    def search(self):
        """
        Description:
            Attempts to identify the string location based on the address passed in the constructor.

            A reminder: abc doesn't care about your parameters matching this prototype, just the name,
            so using **kwargs on search is fine. However, because find_encoded_strings is calling
            this function, there may be no *args as it wouldn't know what to pass. Assuming you don't
            edit this file's code, it may be useful to simply wrap your actual search function with this
            search function in some cases.

        Output:
            True if the encoded string offset is identified, False if the encoded string offset
            is not identified.
        """
        pass


def get_encoded_strings():
    """ Use responsibly """
    return ENCODED_STRINGS


def get_decoded_strings():
    """ Use responsibly """
    return DECODED_STRINGS


def is_valid_ea(ea):
    """
    Description:
        Returns true for valid EAs, False for invalid ones.

    Input:
        ea - The EA to check

    Output:
        True if the EA is valid, False if it is not
    """
    return ea not in (None, idc.BADADDR) and idc.get_inf_attr(
        idc.INF_MIN_EA
    ) <= ea <= idc.get_inf_attr(idc.INF_MAX_EA)


# TODO: Integrate this into EncodedString class
def split_decoded_string(decoded_string, identify=False):
    """
    Description:
        Given a single EncodedString, split it into multiple.
        By default, it is split on \x00.
        The original EncodedString is not modified.

    Input:
        decoded_string - The EncodedString to split
        identify - When True, define the resultant strings in the IDB

    Output:
        A list of EncodedStrings.
    """
    """
    Normally, we have either a block of either all one byte characters OR all two byte characters,
    however with some families, the blocks have utf-8 and utf-16 mixed together. When the block is all one
    byte characters, we just split the block on \x00 and are done with it. Likewise, when the block is all
    two byte characters, we just split the block on \x00\x00 and are done with it. For mixed blocks,
    we can't just split on one or the other. If we were to split on \x00 to find the single byte character strings,
    we would of course completely destroy the two byte character strings. Therefore, we effectively tokenize
    the block on \x00\x00 and process each sub-block separately.

    To process a block, we start from its end (which is guaranteed to be either \x00\x00 or the end of
    the buffer, which might as well be \x00) and work our way forward. We initially start by finding
    the rightmost \x00, which is either the end of the previous single byte character string or the
    first byte in a two byte character. To determine if the \x00 is part of two byte character string,
    we step back two characters. We keep stepping back as long as we keep landing
    on \x00. Once we hit a character that's not \x00, we undo our last step (and because we don't want
    to grab a leading \x00, step one more character, for a total of 3 characters not the 2 we have
    been stepping).

    Wherever we are now is the beginning of the string, which continues up until the index from which
    we searched for the rightmost \x00 earlier. We continue like this until we run into the front of
    the current block and then process the next block, repeating until we hit the end of the buffer.
    """
    results = []
    section_start = 0
    section_end = decoded_string.decoded_data.find(b"\x00\x00")
    if section_end == -1:
        section_end = decoded_string.size
    string_end = section_end

    while True:
        while True:
            # Determine where the next individual string starts
            string_start = decoded_string.decoded_data[section_start:string_end].rfind(
                b"\x00"
            )
            if string_start != -1:
                string_start += section_start
                while (
                    string_start >= section_start
                    and decoded_string.decoded_data[string_start] == 0
                ):
                    string_start -= 2
                if string_start > section_start:
                    string_start += 3  # last step +1 to skip \x00
                elif string_start < section_start:
                    # The leftmost string in the section, so don't step past section_start
                    string_start = section_start
            else:
                string_start = section_start  # The leftmost string in the block

            # Now that we have a string_start and string_end, we can carve the string
            new_decoded_string = decoded_string.decoded_data[string_start:string_end]
            if new_decoded_string:
                # Since we're using \x00 as the delimiter, we need to add \x00 to the end
                new_decoded_string += b"\x00"
                new_string = copy.copy(decoded_string)
                new_string.decoded_data = new_decoded_string
                new_string.size = len(new_decoded_string)
                new_string.offset = string_start
                results.append(new_string)

            if string_start == section_start:
                break
            else:
                # -1 to skip \x00 for searching (which necessitates adding it back above)
                string_end = string_start - 1

        # We've processed a full section, advance to the next section
        if section_end == decoded_string.size:
            break
        else:
            section_start = section_end + 2
            # Skip blocks of \x00
            while (
                section_start < decoded_string.size
                and decoded_string.decoded_data[section_start] == 0
            ):
                section_start += 1
            section_end = decoded_string.decoded_data[section_start:].find(b"\x00\x00")
            if section_end == -1:  # The rightmost section in the block
                section_end = decoded_string.size
            else:
                section_end += section_start
            string_end = section_end

    results.sort(key=lambda decoded_string: decoded_string.start_ea)

    if identify:
        for new_string in results:
            new_string.define()

    return results


def find_encoded_strings_inline(matches, Tracer, **kwargs):
    """
    For each yara match, attempt to find encoded strings.

    Input:
        matches - A list of yara matches (ea, identifier)
        Tracer - A pointer to an implementation of StringTracer.
        **kwargs - kwargs to be passed to Tracer's constructor.

    Output:
        A list of EncodedStrings.
    """
    encoded_strings = []
    for ea, identifier in matches:
        try:
            tracer = Tracer(ea, identifier, **kwargs)
            if tracer.search():
                encoded_strings.extend(tracer.encoded_strings)
            else:
                logger.warning("Failed to find strings at 0x%X" % ea)
        except AttributeError:
            logger.warning("Error tracing at 0x%X" % ea)
    return encoded_strings


def find_encoded_strings(funcs, Tracer, **kwargs):
    """
    Description:
        For each ref for each function, attempt to find encoded strings.

    Input:
        funcs - A list of functions. Must have an xrefs_to field.
        Tracer - A pointer to an implementation of StringTracer.
        **kwargs - kwargs to be passed to Tracer's constructor.

    Output:
        A list of EncodedStrings.
    """
    encoded_strings = []
    for func in funcs:
        for ref in func.xrefs_to:
            if idc.get_segm_name(ref) == ".pdata":
                logger.info(
                    "Segment .pdata for ref 0x%08x is not a relevant code segment and will be skipped"
                    % ref
                )
            else:
                # NOTE: Setting errors to info because it is common and will spam our console.
                try:
                    tracer = Tracer(ref, func.identifier, **kwargs)
                    if tracer.search():
                        encoded_strings.extend(tracer.encoded_strings)
                    else:
                        logger.info("Failed to find strings at 0x%X" % ref)
                except AttributeError as e:
                    # Only catch AttributeErrors resulting from a function not existing. All other AttributeErrors
                    # are actual errors and should go uncaught.
                    #
                    # TODO: Create a separate exception class for a function not existing, so that we don't have to do
                    #  this kind of error message checking.
                    if str(e).startswith("No function at 0x"):  #
                        logger.info(
                            "No function exists at 0x%X. Create a function at this location to obtain strings."
                            % ref
                        )
                    else:
                        raise e
    return encoded_strings


def decode_strings(encoded_strings, decode):
    """
    Description:
        For each encoded_string entry in encoded_strings, decode the data using the provided
        decode function.

    Input:
        encoded_strings - A list of EncodedStrings.
        decode - A pointer to a function that handles decoding.

    Output:
        A list of successfully decoded EncodedStrings.
    """
    # NOTE: Setting errors to info because it is common and will spam our console.
    decoded_strings = []
    for encoded_string in encoded_strings:
        if not encoded_string.encoded_data:
            logger.info("Unable to find string {!r}".format(encoded_string))
            continue
        encoded_string.decoded_data = decode(encoded_string)
        if isinstance(encoded_string.decoded_data, bytearray):
            encoded_string.decoded_data = bytes(encoded_string.decoded_data)
        if encoded_string.decoded_data:  # Allow decoders to abort/fail quietly
            decoded_strings.append(encoded_string)
        else:
            logger.info("Failed to decode string: {!r}".format(encoded_string))
    return decoded_strings


def yara_find_decode_functions(rule_text, func_name=None):
    """
    Description:
        Use yara to find the string decode functions, rename them, and return the SuperFunc_ts.
        Clear the matches each time to prevent duplicates.

    Input:
        rule-text - A string containing a YARA rule
        func_name - The name to be applied to the found function(s). No name will be applied
                    if func_name = None.

    Output:
        A list of SuperFunc_t objects.

    Throws:
        RuntimeError - Assumes that there's no point in continuing if there is no YARA match
                       and that we were expecting a YARA match, so error in that case.
    """
    rule = yara.compile(source=rule_text)
    matches = rule.match_strings()

    if not matches:
        raise RuntimeError(
            "All yara rules failed to match: {}".format(",".format(rule.names))
        )
    return make_superfunc_t_from_matches(matches, func_name)


# TODO: Deprecate this.
def make_superfunc_t_from_matches(matches, func_name=None):
    """
    Description:
        Makes a SuperFunc_t object for each yara match

    Input:
        matches - A list of (ea, identifier) all of which are inside a function.
        func_name - If a non-None name is provided, rename the functions with that name.

    Output:
        A list of decoded functions. Renames those functions if a name is provided.
    """
    decode_funcs = set()
    for ea, identifier in matches:
        if ea == idc.BADADDR:
            continue
        func = SuperFunc_t(ea, identifier)
        if func_name is not None:
            func.rename(func_name)
        decode_funcs.add(func)
    return list(decode_funcs)


def re_find_functions(regex, flags=0, section=None, func_name=None):
    """
    Description:
        Use IDA_re to find the string decode functions, rename them, and return the SuperFunc_ts.
        Clear the matches each time to prevent duplicates.

    Input:
        regex - A compiled regular expression or regular expression string or an ida_re.Pattern object
        section - PE section to restrict searches to
        func_name - The name to be applied to the found function(s). No name will be applied
                    if func_name = None.

    Output:
        A list of SuperFunc_t objects.

    Throws:
        RuntimeError - Assumes that there's no point in continuing if there is no YARA match
                       and that we were expecting a YARA match, so error in that case.
    """
    if isinstance(regex, ida_re.Pattern):
        _regex = regex
    else:
        _regex = ida_re.Pattern(regex, flags)

    funcs = set()
    for match in _regex.finditer(section):
        try:
            func = SuperFunc_t(match.start())
        except AttributeError:
            continue

        if func_name is not None:
            func.rename(func_name)

        funcs.add(func)

    return list(funcs)


def find_input_file():
    """
    Description:
        Check whether or not IDA knows where the original file used to create the IDB is.
        If IDA doesn't know, check the IDA's directory for the file.

    Output:
        Returns True if the input file was located, False if it was not.
    """
    global INPUT_FILE_PATH
    ida_path = INPUT_FILE_PATH
    if not os.path.exists(ida_path):
        # If IDA does not know, check if the (correct) file is sitting next to the IDB.
        local_path = os.path.join(idautils.GetIdbDir(), idc.get_root_filename())
        if (
            os.path.exists(local_path)
            and hashlib.md5(open(local_path, "rb").read()).hexdigest().upper()
            == idc.retrieve_input_file_md5()
        ):
            INPUT_FILE_PATH = local_path
            logger.debug("Guessed the input file path: " + INPUT_FILE_PATH)
            logger.debug("IDA thought it was:          " + ida_path)
            return True
        else:
            return False
    else:
        return True


def string_decoder_main(
    yara_rule,
    Tracer,
    decode,
    patch=True,
    func_name="string_decode_function",
    inline=False,
):
    """
    Description:
        If you are going to use this file's workflow as is, this is the entry point. This supports the majority
        of decode function xref based decoding algorithms (the 'older' way) and inline decoding algorithms
        (the 'newer' way).

    Input:
        First, you'll need to provide a YARA rule. The rule can either be passed as a string or in a file.
        If it is in a file, be sure to set is_yara_file = True. Also note that when YARA finds the decoder
        function, it will attempt to rename it:
            func_name = '' will reset the name to the default name it had when IDA built the IDB
            func_name = None will prevent this script from renaming it

        Second, you need to implement the StringTracer class above to handle your malware's particular
        way of loading the encrypted string. Pass a pointer to this class as the second parameter.

        Third, you need to implement the decoding algorithm and pass a pointer to the entry function
        as the third parameter.

        It is also important to have the original file in the correct location for YARA rule application.
        The IDB contains the path of original file from when the IDB was created, but if the file isn't there,
        this script will check the IDB's current directory for a file with the same name and same md5.

        This script will also patch the encoded string with the decoded one unless patch = False.

    Throws:
        AttributeError - see StringTracer, SuperFunc_t
        OSError - The provided YARA file could not be opened.
        RuntimeError - see _yara_find_decode_functions
        TypeError - The provided Tracer does not extend StringTracer.
        ValueError - Could not find the file used to create the IDB.
                   - see EncodedString.get_bytes
    """
    global ENCODED_STRINGS
    global DECODED_STRINGS

    # We could just check if there is a 'search' method, but handle it this way to enforce conventions.
    if not issubclass(Tracer, StringTracer):
        logger.error("Tracer does not extend StringTracer!")
        return

    # Check that IDA actually knows where the original input file is.
    if not find_input_file():
        logger.error(
            "Unable to locate the file used to create the IDB: " + INPUT_FILE_PATH
        )
        return

    # Do the decoding.
    try:
        if inline:
            matches = yara.match_strings(yara_rule)
            ENCODED_STRINGS = find_encoded_strings_inline(matches, Tracer)
        else:
            decode_functions = yara_find_decode_functions(yara_rule, func_name)
            ENCODED_STRINGS = find_encoded_strings(decode_functions, Tracer)
        ENCODED_STRINGS = decode_strings(ENCODED_STRINGS, decode)
        ENCODED_STRINGS = sorted(set(ENCODED_STRINGS))
        for string in ENCODED_STRINGS:
            string.publish(patch=patch)
        return ENCODED_STRINGS
    except RuntimeError as e:
        logger.error(
            "The provided YARA rule failed to match. No strings can be decrypted for this YARA rule."
        )
        return


idc.auto_wait()  # Force wait on import just to be sure
